####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2017 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

"""This module implements a parser for Spice expressions.
"""

####################################################################################################

import logging

####################################################################################################

import ply.lex as lex
import ply.yacc as yacc

####################################################################################################

from .Ast import *

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

# def ensure_statement_list(x):
#     if isinstance(x, StatementList):
#         return x
#     else:
#         return StatementList(x)

####################################################################################################

class Parser:

    _logger = _module_logger.getChild('Parser')

    ##############################################

    reserved = {
    }

    tokens = [
        'NAME',
        # 'INT', 'FLOAT',
        'NUMBER',
        'SEMICOLON',
        'LEFT_PARENTHESIS', 'RIGHT_PARENTHESIS',
        'SET',
        'NOT',
        'POWER',
        'MULTIPLY',
        'DIVIDE', 'INT_DIVIDE', 'MODULO',
        'PLUS', 'MINUS',
        'EQUAL', 'NOT_EQUAL', 'LESS', 'GREATER', 'LESS_EQUAL', 'GREATER_EQUAL',
        'AND', 'OR',
        'IF', 'COLON',
    ] + list(reserved.values())

    ##############################################

    def t_error(self, token):
        self._logger.error("Illegal character '%s' at line %u and position %u" %
                           (token.value[0],
                            token.lexer.lineno,
                            token.lexer.lexpos - self._previous_newline_position))
        # token.lexer.skip(1)
        raise NameError('Lexer error')

    ##############################################

    t_ignore  = ' \t'

    def t_newline(self, t):
        r'\r?\n+'
        # Track newline
        t.lexer.lineno += len(t.value)
        self._previous_newline_position = t.lexer.lexpos
        # t.type = 'SEMICOLON'
        # return t

    t_ignore_COMMENT = r'\#[^\n]*'

    ##############################################

    t_SEMICOLON = r';'

    t_LEFT_PARENTHESIS = r'\('
    t_RIGHT_PARENTHESIS = r'\)'

    t_SET = r'='

    t_NOT = r'!'

    t_POWER = r'\*\*'

    t_MULTIPLY = r'\*'
    t_DIVIDE = r'/'
    t_MODULO = r'%'
    t_INT_DIVIDE = r'\/'

    t_PLUS = r'\+'
    t_MINUS = r'-'

    t_EQUAL = r'=='
    t_NOT_EQUAL = r'!='
    t_LESS = r'<'
    t_GREATER = r'>'
    t_LESS_EQUAL = r'<='
    t_GREATER_EQUAL = r'>='

    t_AND = r'&&'
    t_OR = r'\|\|'

    t_IF = r'\?'
    t_COLON = r':'

    def t_NAME(self, t):
        r'[a-zA-Z_][a-zA-Z_0-9]*'
        # Check for reserved words
        t.type = self.reserved.get(t.value, 'NAME') # Fixme: ???
        return t

    # def t_INT(self, t):
    #     r'\d+'
    #     t.value = int(t.value)
    #     return t

    # exponent_part = r"""([eE][-+]?[0-9]+)"""
    # fractional_constant = r"""([0-9]*\.[0-9]+)|([0-9]+\.)"""
    # floating_constant = '(((('+fractional_constant+')'+exponent_part+'?)|([0-9]+'+exponent_part+'))[FfLl]?)'

    def t_NUMBER(self, t):
        # 1 1. 1.23 .1
        # r'\d*\.?\d+([eE][-+]?\d+)?'
        # r'(\d+)(\.\d+)(e(\+|-)?(\d+))? | (\d+)e(\+|-)?(\d+)'
        r'\d+\.\d+(e(\+|-)?(\d+))? | \d+\.(e(\+|-)?(\d+))? | \.\d+(e(\+|-)?(\d+))? | \d+'
        t.value = t.value
        return t

    ##############################################
    #
    # Grammar
    #

    # from lowest
    precedence = (
        ('left', 'IF'),
        ('left', 'OR'),
        ('left', 'AND'),
        ('left', 'GREATER', 'LESS', 'GREATER_EQUAL', 'LESS_EQUAL', 'NOT_EQUAL', 'EQUAL'),
        ('left', 'MINUS', 'PLUS'),
        ('left', 'INT_DIVIDE', 'MODULO', 'DIVIDE', 'MULTIPLY'),
        ('left', 'POWER'),
        ('left', 'NOT'), # , 'NEGATION'
    )

    def p_error(self, p):
        if p:
            self._logger.error("Syntax error at '%s'", p.value)
            raise NameError('Syntax Error')
        else:
            self._logger.error("Syntax Error at End Of File")
            raise NameError("Syntax Error at End Of File" )

    # start = 'program'
    start = 'statement'

    # def p_empty(self, p):
    #     'empty :'
    #     pass

    def p_statement(self, t):
        'statement : expression'
        print('statement', t[1])

    # def p_program(self, p):
    #     '''program : statement
    #                | program statement
    #                | empty
    #     '''
    #     if len(p) == 3:
    #         statement = p[2]
    #     else:
    #         statement = p[1]
    #     if statement is not None:
    #         self._program.add(statement)

    # def p_statement(self, p):
    #     '''statement : expression_statement
    #     '''
    #     p[0] = p[1]

    # def p_expression_statement(self, p):
    #     '''expression_statement : assignation SEMICOLON
    #                             | function SEMICOLON
    #                             | SEMICOLON
    #     '''
    #     if len(p) == 3:
    #         p[0] = p[1]

    # def p_statement_list(self, p):
    #     '''statement_list : statement
    #                       | statement_list statement
    #     '''
    #     if len(p) == 3:
    #         p[1].add(p[2])
    #         p[0] = p[1]
    #     else:
    #         p[0] = StatementList(p[1])

    # def p_expression_list(self, p):
    #     '''expression_list : expression
    #                        | expression_list COMMA expression
    #     '''
    #     if len(p) == 3:
    #         p[1].add(p[2])
    #         p[0] = p[1]
    #     else:
    #         p[0] = StatementList(p[1])

    # def p_function(self, p):
    #     '''function : NAME LEFT_PARENTHESIS expression_list RIGHT_PARENTHESIS
    #                 | NAME LEFT_PARENTHESIS RIGHT_PARENTHESIS
    #     '''
    #     if len(p) == 5:
    #         p[0] = Function(p[1], p[3])
    #     else:
    #         p[0] = Function(p[1])

    def p_variable(self, p):
        '''variable : NAME
        '''
        p[0] = Variable(p[1])

    # def p_assignation(self, p):
    #     'assignation : variable SET expression'
    #     p[0] = Assignation(p[3], p[1]) # eval value first

    # def p_interger(self, p):
    #     '''constant : INT
    #     '''
    #     p[0] = IntConstant(p[1])

    def p_float(self, p):
        '''constant : NUMBER
        '''
        if '.' in p[1]:
            p[0] = FloatConstant(p[1])
        else:
            p[0] = IntConstant(p[1])

    def p_value(self, p):
        '''expression : variable
                      | constant
        '''
        p[0] = p[1]

    def p_unnary_operation(self, p):
        # OP ...
        '''expression : MINUS expression
                      | NOT expression
        '''
        p[0] = OperatorMetaclass.get_unary(p[1])(p[2])

    def p_binary_operation(self, p):
        # ... OP ...
        '''expression : expression POWER expression
                      | expression MULTIPLY expression
                      | expression DIVIDE expression
                      | expression MODULO expression
                      | expression INT_DIVIDE expression
                      | expression PLUS expression
                      | expression MINUS expression
                      | expression EQUAL expression
                      | expression NOT_EQUAL expression
                      | expression LESS expression
                      | expression GREATER expression
                      | expression LESS_EQUAL expression
                      | expression GREATER_EQUAL expression
                      | expression AND expression
                      | expression OR expression
        '''
        p[0] = OperatorMetaclass.get_binary(p[2])(p[1], p[3])

    def p_if(self, p):
        '''expression : expression IF expression COLON expression
        '''
        p[0] = If(p[1], p[3], p[5])

    ##############################################

    def __init__(self):

        self._build()

    ##############################################

    def _build(self, **kwargs):

        self._lexer = lex.lex(module=self, **kwargs)
        self._parser = yacc.yacc(module=self, **kwargs)

    ##############################################

    def _reset(self):

        self._previous_newline_position = 0
        # self._program = Program()

    ##############################################

    def parse(self, text):

        self._reset() # Fixme: after ?
        self._parser.parse(text, lexer=self._lexer)
        # return self._program

    ##############################################

    def test_lexer(self, text):

        self._reset()
        self._lexer.input(text)
        while True:
            token = self._lexer.token()
            if not token:
                break
            print(token)



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

import unittest

import numpy as np

from scipy.special import gamma
from scipy.linalg import solve

####################################################################################################

from PySpice.Math import odd
from PySpice.Math.Calculus import compute_exact_finite_difference_coefficients, derivative

####################################################################################################

# From Generation of Finite Difference Formulas on Arbitrary Space Grids
#   Bengt Fornberg, Mathematics of computation, volume 51, number 184, october 1988

centred_coefficients = {
    # on a grid -4,...,4
    # [order of derivative][order of accuracy] = (left coefficients, ...)
    0: { None: [1,] },
    1: {
        2: [-1/2.,   0],
        4: [1/12.,   -2/3.,   0],
        6: [-1/60.,   3/20.,   -3/4.,   0],
        8: [1/280.,   -4/105.,   1/5.,   -4/5.,   0],
   },
    2: {
        2: [1,  -2],
        4: [-1/12.,  4/3.,  -5/2.],
        6: [1/90.,  -3/20.,  3/2.,  -49/18.],
        8: [-1/560.,  8/315.,  -1/5.,  8/5.,  -205/72.],
   },
}
# complete right coefficients
for derivative_order, derivative_order_dict in centred_coefficients.items():
    for coefficients in derivative_order_dict.values():
        if len(coefficients) > 1:
            for i in range(len(coefficients) -2, -1, -1):
                coefficient = coefficients[i]
                if odd(derivative_order):
                    coefficient *= -1
                coefficients.append(coefficient)

forward_coefficients = {
    # on a grid 0...8
    # [order of derivative][order of accuracy] = (coefficients)
    0: { None:(1,) },
    1: {
        1: (-1,  1),
        2: (-3/2.,  2,  -1/2.),
        3: (-11/6.,  3,  -3/2.,  1/3.),
        4: (-25/12.,  4,  -3,  4/3.,  -1/4.),
        5: (-137/60.,  5,  -5,  10/3.,  -5/4.,  1/5.),
        6: (-49/20.,  6,  -15/2.,  20/3.,  -15/4.,  6/5.,  -1/6.),
        7: (-363/140.,  7,  -21/2.,  35/3.,  -35/4.,  21/5.,  -7/6.,  1/7.),
        8: (-761/280.,  8,  -14,  56/3.,  -35/2.,  56/5.,  -14/3.,  8/7.,  -1/8.),
   },
    2: {
        1: (1,  -2,  1),
        2: (2,  -5,  4,  -1),
        3: (35/12.,  -26/3.,  19/2.,  -14/3.,  11/12.),
        4: (15/4.,  -77/6.,  107/6.,  -13,  61/12.,  -5/6.),
        5: (203/45.,  -87/5.,  117/4.,  -254/9.,  33/2.,  -27/5.,  137/180.),
        6: (469/90.,  -223/10.,  879/20.,  -949/18.,  41,  -201/10.,  1019/180.,  -7/10.),
        7: (29531/5040.,  -962/35.,  621/10.,  -4006/45.,   691/8.,  -282/5.,  2143/90.,  -206/35.,  363/560.),
   },
}

####################################################################################################

def compute_finite_difference_coefficients(derivative_order, grid_size):

    # from http://www.scientificpython.net/pyblog/uniform-finite-differences-all-orders

    n = 2*grid_size -1
    A = np.tile(np.arange(grid_size), (n,1)).T
    B = np.tile(np.arange(1-grid_size,grid_size), (grid_size,1))
    M = (B**A)/gamma(A+1)

    r = np.zeros(grid_size)
    r[derivative_order] = 1

    D = np.zeros((grid_size, grid_size))
    for k in range(grid_size):
        indexes = k + np.arange(grid_size)
        D[:,k] = solve(M[:,indexes], r)

    return D

####################################################################################################

class TestFiniteDifference(unittest.TestCase):

    ##############################################

    def test_coefficient(self):

        for derivative_order, derivative_order_dict in centred_coefficients.items():
            for accuracy_order, coefficients in derivative_order_dict.items():
                if accuracy_order is not None and accuracy_order > 1:
                    coefficients = np.array(coefficients)
                    grid_size = accuracy_order +1
                    # D = compute_finite_difference_coefficients(derivative_order, grid_size)
                    # computed_coefficients = D[:,grid_size/2]
                    grid = list(range(-accuracy_order//2,accuracy_order//2+1))
                    computed_coefficients = compute_exact_finite_difference_coefficients(derivative_order, grid)
                    computed_coefficients = [float(x) for x in computed_coefficients]
                    # print "Derivative order {} Accuracy order {}".format(derivative_order, accuracy_order)
                    # print coefficients
                    # print computed_coefficients
                    # print np.abs(coefficients - computed_coefficients)
                    self.assertTrue(np.all(np.isclose(coefficients, computed_coefficients)))

        for derivative_order, derivative_order_dict in forward_coefficients.items():
            for accuracy_order, coefficients in derivative_order_dict.items():
                if accuracy_order is not None and accuracy_order > 1:
                    coefficients = np.array(coefficients)
                    grid_size = derivative_order + accuracy_order
                    # D = compute_finite_difference_coefficients(derivative_order, grid_size)
                    # computed_coefficients = D[:,-1]
                    grid = list(range(grid_size))
                    computed_coefficients = compute_exact_finite_difference_coefficients(derivative_order, grid)
                    computed_coefficients = [float(x) for x in computed_coefficients]
                    # print "Derivative order {} Accuracy order {}".format(derivative_order, accuracy_order)
                    # print D
                    # print coefficients
                    # print computed_coefficients
                    # print np.abs(coefficients - computed_coefficients)
                    self.assertTrue(np.all(np.isclose(coefficients, computed_coefficients)))

    ##############################################

    def test_derivative(self):

        x = np.linspace(0, 2*np.pi, 100)
        y = np.sin(x)
        true_dy1 = np.cos(x) # d sin = cos
        true_dy2 = -y # d cos = - sin

        dy1 = derivative(x, y, derivative_order=1, accuracy_order=4)
        dy2 = derivative(x, y, derivative_order=2, accuracy_order=4)

        # import matplotlib.pyplot as plt
        # plt.plot(x[:-1], np.diff(y)/np.diff(x) - true_dy1[:-1])
        # plt.plot(x, dy1 - true_dy1)
        # plt.plot(x, dy2 - true_dy2)
        # plt.show()
        np.testing.assert_array_almost_equal(dy1, true_dy1, decimal=1)
        np.testing.assert_array_almost_equal(dy2, true_dy2, decimal=1)

####################################################################################################

if __name__ == '__main__':

    unittest.main()



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2017 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

"""This module provides an interface to run xyce and get back the simulation
output.

"""

####################################################################################################

import logging
import os
import shutil
import subprocess
import tempfile

from PySpice.Config import ConfigInstall
from .RawFile import RawFile

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class XyceServer:

    """This class wraps the execution of Xyce and convert the output to a Python data structure.

    Example of usage::

      spice_server = XyceServer(xyce_command='/path/to/Xyce')
      raw_file = spice_server(spice_input)

    It returns a :obj:`PySpice.Spice.RawFile` instance.

    Default Xyce path is set in `XyceServer.XYCE_COMMAND`.

    """

    if ConfigInstall.OS.on_linux:
        XYCE_COMMAND = 'Xyce'
    elif ConfigInstall.OS.on_osx:
        XYCE_COMMAND = 'Xyce'
    elif ConfigInstall.OS.on_windows:
        XYCE_COMMAND = 'C:\\Program Files\\Xyce 6.10 OPENSOURCE\\bin\\Xyce.exe'
    else:
        raise NotImplementedError

    _logger = _module_logger.getChild('XyceServer')

    ##############################################

    def __init__(self, **kwargs):

        self._xyce_command = kwargs.get('xyce_command') or self.XYCE_COMMAND

    ##############################################

    def _parse_stdout(self, stdout):

        """Parse stdout for errors."""

        # log Spice output
        self._logger.info(os.linesep + stdout.decode('utf-8'))

        error_found = False
        simulation_failed = False
        warning_found = False
        lines = stdout.splitlines()
        for line_index, line in enumerate(lines):
            if line.startswith(b'Netlist warning'):
                warning_found = True
                # Fixme: highlight warnings
                self._logger.warning(os.linesep + line.decode('utf-8'))
            elif line.startswith(b'Netlist error'):
                error_found = True
                self._logger.error(os.linesep + line.decode('utf-8'))
            elif b'Transient failure history' in line:
                simulation_failed = True
                self._logger.error(os.linesep + line.decode('utf-8'))
        if error_found:
            raise NameError("Errors was found by Xyce")
        elif simulation_failed:
            raise NameError("Xyce simulation failed")

    ##############################################

    def __call__(self, spice_input):

        """Run SPICE in server mode as a subprocess for the given input and return a
        :obj:`PySpice.RawFile.RawFile` instance.

        """

        self._logger.debug('Start the xyce subprocess')

        tmp_dir = tempfile.mkdtemp()
        input_filename = os.path.join(tmp_dir, 'input.cir')
        output_filename = os.path.join(tmp_dir, 'output.raw')
        with open(input_filename, 'w') as f:
            f.write(str(spice_input))

        command = (self._xyce_command, '-r', output_filename, input_filename)
        self._logger.info('Run {}'.format(' '.join(command)))
        process = subprocess.Popen(
            command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = process.communicate()

        self._parse_stdout(stdout)

        with open(output_filename, 'rb') as f:
            output = f.read()
        # self._logger.debug(output)

        raw_file = RawFile(output)
        shutil.rmtree(tmp_dir)

        return raw_file



####################################################################################################
#
# PySpice - A Spice package for Python
# Copyright (C) 2019 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

import datetime
fromtimestamp = datetime.datetime.fromtimestamp

from invoke import task
# import sys

try:
    import pygit2 as git
except ImportError:
    git = None

####################################################################################################

@task
def push_tags(ctx):
    # git push --delete origin v1.4.3_beta0
    # git tag -d v1.4.3_beta0
    ctx.run('git push --tags')

####################################################################################################

@task
def uncommit(ctx):
    ctx.run('git reset --soft HEAD~')

####################################################################################################

@task
def get_commits(ctx):

    path = '.'
    repository_path = git.discover_repository(path)
    # print(repository_path)
    repository = git.Repository(repository_path)

    head = repository.head
    head_commit = repository[head.target]

    # GIT_SORT_TOPOLOGICAL. Sort the repository contents in topological order (parents before children);
    # this sorting mode can be combined with time sorting.
    sorting = git.GIT_SORT_TOPOLOGICAL   # git.GIT_SORT_TIME
    commits = [commit for commit in repository.walk(head_commit.id, sorting)]   # Fixme:

    template = '''
====================================================================================================
{message}

  -------------------------------------------------
  {name} <{email}> / {timestamp}
  {id}
'''

    for commit in commits:
        fields = {
            'message': commit.message.strip(),
            'id': commit.hex,
            'timestamp': fromtimestamp(commit.commit_time).strftime('%Y-%m-%d %H:%M:%S'),
            'name': commit.committer.name,   # author
            'email': commit.committer.email,
        }
        if 'salvaire' not in fields['name'].lower():
            print(template.lstrip().format(**fields))

####################################################################################################

@task
def fetch_pr(ctx, pr_id):
    # https://docs.github.com/en/github/collaborating-with-issues-and-pull-requests/checking-out-pull-requests-locally
    branch_name = f'pr/{pr_id}'
    ctx.run(f'git fetch origin pull/{pr_id}/head:{branch_name}')
    # git checkout {branch_name}
    # rework ...
    # git push origin {branch_name}
    # -> create new PR

# @task
# def fetch_pr_range(ctx, start_id, stop_id):
#     for i in range(int(start_id), int(stop_id+1)):
#         ctx.run('git fetch origin pull/{0}/head:pr/{0}'.format(i))

####################################################################################################

#  https://docs.github.com/en/github/collaborating-with-issues-and-pull-requests/merging-a-pull-request



####################################################################################################

#r#
#r# =====================
#r#  NgSpice Interpreter
#r# =====================
#r#
#r# This example explains how to use the NgSpice binding.
#r#

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Spice.NgSpice.Shared import NgSpiceShared

####################################################################################################

ngspice = NgSpiceShared.new_instance()

print(ngspice.exec_command('version -f'))
print(ngspice.exec_command('print all'))
print(ngspice.exec_command('devhelp'))
print(ngspice.exec_command('devhelp resistor'))

circuit = '''
.title Voltage Multiplier

.SUBCKT 1N4148 1 2
*
R1 1 2 5.827E+9
D1 1 2 1N4148
*
.MODEL 1N4148 D
+ IS = 4.352E-9
+ N = 1.906
+ BV = 110
+ IBV = 0.0001
+ RS = 0.6458
+ CJO = 7.048E-13
+ VJ = 0.869
+ M = 0.03
+ FC = 0.5
+ TT = 3.48E-9
.ENDS

Vinput in 0 DC 0V AC 1V SIN(0V 10V 50Hz 0s 0Hz)
C0 in 1 1mF
X0 1 0 1N4148
C1 0 2 1mF
X1 2 1 1N4148
C2 1 3 1mF
X2 3 2 1N4148
C3 2 4 1mF
X3 4 3 1N4148
C4 3 5 1mF
X4 5 4 1N4148
R1 5 6 1MegOhm
.options TEMP = 25°C
.options TNOM = 25°C
.options filetype = binary
.options NOINIT
.ic
.tran 0.0001s 0.4s 0s
.end
'''

ngspice.load_circuit(circuit)
print('Loaded circuit:')
print(ngspice.listing())

print(ngspice.show('c3'))
print(ngspice.showmod('c3'))

ngspice.run()
print('Plots:', ngspice.plot_names)

print(ngspice.ressource_usage())
print(ngspice.status())

plot = ngspice.plot(simulation=None, plot_name=ngspice.last_plot)
print(plot)

# ngspice.quit()



####################################################################################################

#r#
#r# =================================================
#r#  Three-phased Current: Y and Delta configurations
#r# =================================================
#r#
#r# This examples shows the computation of the voltage for the Y and Delta configurations.
#r#

####################################################################################################

import math

import numpy as np
import matplotlib.pyplot as plt

####################################################################################################

from PySpice.Unit import *

####################################################################################################

#r# Let use an European 230 V / 50 Hz electric network.

frequency = 50@u_Hz
w = frequency.pulsation
period = frequency.period

rms_mono = 230
amplitude_mono = rms_mono * math.sqrt(2)

#r# The phase voltages in Y configuration are dephased of :math:`\frac{2\pi}{3}`:
#r#
#r# .. math::
#r#  V_{L1 - N} = V_{pp} \cos \left( \omega t \right) \\
#r#  V_{L2 - N} = V_{pp} \cos \left( \omega t - \frac{2\pi}{3} \right) \\
#r#  V_{L3 - N} = V_{pp} \cos \left( \omega t - \frac{4\pi}{3} \right)
#r#
#r# We rewrite them in complex notation:
#r#
#r# .. math::
#r#  V_{L1 - N} = V_{pp} e^{j\omega t} \\
#r#  V_{L2 - N} = V_{pp} e^{j \left(\omega t - \frac{2\pi}{3} \right) } \\
#r#  V_{L3 - N} = V_{pp} e^{j \left(\omega t - \frac{4\pi}{3} \right) }

t = np.linspace(0, 3*float(period), 1000)
L1 = amplitude_mono * np.cos(t*w)
L2 = amplitude_mono * np.cos(t*w - 2*math.pi/3)
L3 = amplitude_mono * np.cos(t*w - 4*math.pi/3)

#r# From these expressions, we compute the voltage in delta configuration using trigonometric identities :
#r#
#r# .. math::
#r#   V_{L1 - L2} = V_{L1} \sqrt{3} e^{j \frac{\pi}{6} } \\
#r#   V_{L2 - L3} = V_{L2} \sqrt{3} e^{j \frac{\pi}{6} } \\
#r#   V_{L3 - L1} = V_{L3} \sqrt{3} e^{j \frac{\pi}{6} }
#r#
#r# In comparison to the Y configuration, the voltages in delta configuration are magnified by
#r# a factor :math:`\sqrt{3}` and dephased of :math:`\frac{\pi}{6}`.
#r#
#r# Finally we rewrite them in temporal notation:
#r#
#r# .. math::
#r#  V_{L1 - L2} = V_{pp} \sqrt{3} \cos \left( \omega t + \frac{\pi}{6} \right) \\
#r#  V_{L2 - L3} = V_{pp} \sqrt{3} \cos \left( \omega t - \frac{\pi}{2} \right) \\
#r#  V_{L3 - L1} = V_{pp} \sqrt{3} \cos \left( \omega t - \frac{7\pi}{6} \right)

rms_tri = math.sqrt(3) * rms_mono
amplitude_tri = rms_tri * math.sqrt(2)

L12 = amplitude_tri * np.cos(t*w + math.pi/6)
L23 = amplitude_tri * np.cos(t*w - math.pi/2)
L31 = amplitude_tri * np.cos(t*w - 7*math.pi/6)

#r# Now we plot the waveforms:
figure, ax = plt.subplots(figsize=(20, 10))
ax.plot(
    t, L1, t, L2, t, L3,
    t, L12, t, L23, t, L31,
    # t, L1-L2, t, L2-L3, t, L3-L1,
)
ax.grid()
ax.set_title('Three-phase electric power: Y and Delta configurations (230V Mono/400V Tri 50Hz Europe)')
ax.legend(
    ('L1-N', 'L2-N', 'L3-N',
     'L1-L2', 'L2-L3', 'L3-L1'),
    loc=(.7,.5),
)
ax.set_xlabel('t [s]')
ax.set_ylabel('[V]')
ax.axhline(y=rms_mono, color='blue')
ax.axhline(y=-rms_mono, color='blue')
ax.axhline(y=rms_tri, color='blue')
ax.axhline(y=-rms_tri, color='blue')

plt.show()

#f# save_figure('figure', 'three-phase.png')



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2020 jmgc / Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

from pathlib import Path
import os
import unittest

####################################################################################################

from PySpice.Spice.Netlist import Circuit
from PySpice.Spice.Parser import SpiceParser

####################################################################################################

path = Path(__file__).parent

with open(path.joinpath('hsop77.cir')) as fh:
    hsop77 = fh.read()

with open(path.joinpath('hsada4077.cir')) as fh:
    hsada4077 = fh.read()

####################################################################################################

def circuit_gft(prb):
    circuit_file = SpiceParser(source=prb[0])
    circuit = circuit_file.build_circuit()
    circuit.parameter('prb', str(prb[1]))
    # Fixme: simulate with Xyce, CI !!!
    simulator = circuit.simulator(simulator='xyce-serial')
    simulator.save(['all'])
    return simulator

####################################################################################################

class TestSpiceParser(unittest.TestCase):

    ##############################################

    @unittest.skip('')
    def test_parser(self):
        for source in (hsop77, hsada4077):
            results = list(map(circuit_gft, [(source, -1), (source, 1)]))
            self.assertEqual(len(results), 2)
            values = str(results[0])
            self.assertNotRegex(values, r'(\.ic)')

    ##############################################

    @unittest.skip('')
    def test_subcircuit(self):
        circuit = Circuit('')
        circuit.include('.../mosdriver.lib')
        circuit.X('test', 'mosdriver', '0', '1', '2', '3', '4', '5')
        circuit.BehavioralSource('test', '1', '0', voltage_expression='if(0, 0, 1)', smoothbsrc=1)
        print(circuit)

####################################################################################################

if __name__ == '__main__':
    unittest.main()



#r# This example depicts half and full wave rectification.

####################################################################################################

import matplotlib.pyplot as plt

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Doc.ExampleTools import find_libraries
from PySpice.Probe.Plot import plot
from PySpice.Spice.Library import SpiceLibrary
from PySpice.Spice.Netlist import Circuit
from PySpice.Unit import *

####################################################################################################

libraries_path = find_libraries()
spice_library = SpiceLibrary(libraries_path)

####################################################################################################

figure1, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(20, 10))

####################################################################################################

circuit = Circuit('half-wave rectification')
circuit.include(spice_library['1N4148'])
source = circuit.SinusoidalVoltageSource('input', 'in', circuit.gnd, amplitude=10@u_V, frequency=50@u_Hz)
circuit.X('D1', '1N4148', 'in', 'output')
circuit.R('load', 'output', circuit.gnd, 100@u_Ω)

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.transient(step_time=source.period/200, end_time=source.period*2)

ax1.set_title('Half-Wave Rectification')
ax1.set_xlabel('Time [s]')
ax1.set_ylabel('Voltage [V]')
ax1.grid()
ax1.plot(analysis['in'])
ax1.plot(analysis.output)
ax1.legend(('input', 'output'), loc=(.05,.1))
ax1.set_ylim(float(-source.amplitude*1.1), float(source.amplitude*1.1))

####################################################################################################

#f# circuit_macros('half-wave-rectification.m4')

circuit.C('1', 'output', circuit.gnd, 1@u_mF)

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.transient(step_time=source.period/200, end_time=source.period*2)

ax2.set_title('Half-Wave Rectification with filtering')
ax2.set_xlabel('Time [s]')
ax2.set_ylabel('Voltage [V]')
ax2.grid()
ax2.plot(analysis['in'])
ax2.plot(analysis.output)
ax2.legend(('input', 'output'), loc=(.05,.1))
ax2.set_ylim(float(-source.amplitude*1.1), float(source.amplitude*1.1))

####################################################################################################

circuit = Circuit('half-wave rectification')
circuit.include(spice_library['1N4148'])
source = circuit.SinusoidalVoltageSource('input', 'in', circuit.gnd, amplitude=10, frequency=50)
circuit.X('D1', '1N4148', 'in', 'output_plus')
circuit.R('load', 'output_plus', 'output_minus', 100@u_Ω)
circuit.X('D2', '1N4148', 'output_minus', circuit.gnd)
circuit.X('D3', '1N4148', circuit.gnd, 'output_plus')
circuit.X('D4', '1N4148', 'output_minus', 'in')

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.transient(step_time=source.period/200, end_time=source.period*2)

ax3.set_title('Full-Wave Rectification')
ax3.set_xlabel('Time [s]')
ax3.set_ylabel('Voltage [V]')
ax3.grid()
ax3.plot(analysis['in'])
ax3.plot(analysis.output_plus - analysis.output_minus)
ax3.legend(('input', 'output'), loc=(.05,.1))
ax3.set_ylim(float(-source.amplitude*1.1), float(source.amplitude*1.1))

####################################################################################################

#f# circuit_macros('full-wave-rectification.m4')

circuit.C('1', 'output_plus', 'output_minus', 1@u_mF)

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.transient(step_time=source.period/200, end_time=source.period*2)

ax4.set_title('Full-Wave Rectification with filtering')
ax4.set_xlabel('Time [s]')
ax4.set_ylabel('Voltage [V]')
ax4.grid()
ax4.plot(analysis['in'])
ax4.plot(analysis.output_plus - analysis.output_minus)
ax4.legend(('input', 'output'), loc=(.05,.1))
ax4.set_ylim(float(-source.amplitude*1.1), float(source.amplitude*1.1))

plt.tight_layout()

#f# save_figure('figure1', 'rectification.png')

####################################################################################################

circuit = Circuit('115/230V Rectifier')
circuit.include(spice_library['1N4148'])
on_115 = True # switch to select 115 or 230V
if on_115:
    node_230 = circuit.gnd
    node_115 = 'node_115'
    amplitude = 115@u_V
else:
    node_230 = 'node_230'
    node_115 = circuit.gnd
    amplitude = 230@u_V
source = circuit.SinusoidalVoltageSource('input', 'in', circuit.gnd, amplitude=amplitude, frequency=50) # Fixme: rms
circuit.X('D1', '1N4148', 'in', 'output_plus')
circuit.X('D3', '1N4148', node_230, 'output_plus')
circuit.X('D2', '1N4148', 'output_minus', node_230)
circuit.X('D4', '1N4148', 'output_minus', 'in')
circuit.C('1', 'output_plus', node_115, 1@u_mF)
circuit.C('2', node_115, 'output_minus', 1@u_mF)
circuit.R('load', 'output_plus', 'output_minus', 10@u_Ω)

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
if on_115:
    simulator.initial_condition(node_115=0)
analysis = simulator.transient(step_time=source.period/200, end_time=source.period*2)

figure2, ax = plt.subplots(figsize=(20, 10))
ax.set_title('115/230V Rectifier')
ax.set_xlabel('Time [s]')
ax.set_ylabel('Voltage [V]')
ax.grid()
ax.plot(analysis['in'])
ax.plot(analysis.output_plus - analysis.output_minus)
ax.legend(('input', 'output'), loc=(.05,.1))
# ax.set_ylim(float(-source.amplitude*1.1), float(source.amplitude*1.1))

plt.tight_layout()

#f# save_figure('figure2', 'universal-rectifier.png')

plt.show()



#r# This example depicts a voltage multiplier using diodes and capacitors.  To go further, you can
#r# read this `page <http://en.wikipedia.org/wiki/Voltage_multiplier>`_ on Wikipedia.

####################################################################################################

import matplotlib.pyplot as plt

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Doc.ExampleTools import find_libraries
from PySpice.Probe.Plot import plot
from PySpice.Spice.Library import SpiceLibrary
from PySpice.Spice.Netlist import Circuit
from PySpice.Unit import *

####################################################################################################

libraries_path = find_libraries()
spice_library = SpiceLibrary(libraries_path)

####################################################################################################

#f# circuit_macros('voltage-multiplier-circuit.m4')

circuit = Circuit('Voltage Multiplier')
circuit.include(spice_library['1N4148'])
source = circuit.SinusoidalVoltageSource('input', 'in', circuit.gnd, amplitude=10@u_V, frequency=50@u_Hz)

multiplier = 5
for i in range(multiplier):
    if i:
        top_node = i - 1
    else:
        top_node = 'in'
    midlle_node, bottom_node = i + 1, i
    circuit.C(i, top_node, midlle_node, 1@u_mF)
    circuit.X(i, '1N4148', midlle_node, bottom_node)
circuit.R(1, multiplier, multiplier+1, 1@u_MΩ)

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.transient(step_time=source.period/200, end_time=source.period*20)

####################################################################################################

figure, ax = plt.subplots(figsize=(20, 10))

ax.set_title('Voltage Multiplier')
ax.set_xlabel('Time [s]')
ax.set_ylabel('Voltage [V]')
ax.grid()
# Fixme: axis vs axe ...
ax.plot(analysis['in'])
for i in range(1, multiplier+1):
    y = analysis[str(i)]
    if i & 1: # for odd multiplier the ground is permuted
        y -= analysis['in']
    ax.plot(y)
# ax.axhline(-multiplier*source.amplitude)
ax.set_ylim(float(-multiplier*1.1*source.amplitude), float(1.1*source.amplitude))
ax.legend(['input'] + ['*' + str(i) for i in range(1, multiplier+1)] ,
          loc=(.2,.8))

plt.tight_layout()
plt.show()

#f# save_figure('figure', 'voltage-multiplier.png')



####################################################################################################

#r#
#r# ========================
#r#  Fast Fourier Transform
#r# ========================
#r#
#r# This example shows how to compute a FFT of a signal using the scipy Scientific Python package.
#r#

####################################################################################################

import numpy as np

from scipy import signal
from scipy.fftpack import fft

import matplotlib.pyplot as plt

####################################################################################################

#r#
#r# We will first compute the spectrum of the sum of two sinusoidal waveforms.
#r#

N = 1000 # number of sample points
dt = 1. / 500 # sample spacing

frequency1 = 50.
frequency2 = 80.

t = np.linspace(0.0, N*dt, N)
y = np.sin(2*np.pi * frequency1 * t) + .5 * np.sin(2*np.pi * frequency2 * t)

yf = fft(y)
tf = np.linspace(.0, 1./(2.*dt), N//2)
spectrum = 2./N * np.abs(yf[0:N//2])

figure1, ax = plt.subplots(figsize=(20, 10))

ax.plot(tf, spectrum, 'o-')
ax.grid()
for frequency in frequency1, frequency2:
    ax.axvline(x=frequency, color='red')
ax.set_title('Spectrum')
ax.set_xlabel('Frequency [Hz]')
ax.set_ylabel('Amplitude')

#f# save_figure('figure1', 'fft-sum-of-sin.png')

####################################################################################################

#r#
#r# Now we will compute the spectrum of a square waveform.
#r#
#r# The Fourier series is given by:
#r#
#r# .. math::
#r#
#r#   \frac{4}{\pi} \sum_{n=1, 3, 5, \ldots}^{\inf} \frac{1}{n} \sin(n 2\pi f t)
#r#

N = 1000 # number of sample points
dt = 1. / 1000 # sample spacing

frequency = 5.

t = np.linspace(.0, N*dt, N)
y = signal.square(2*np.pi*frequency*t)

figure2, (ax1, ax2) = plt.subplots(2, figsize=(20, 10))

ax1.plot(t, y)
y_sum = None
for n in range(1, 20, 2):
    yn = 4/(np.pi*n)*np.sin((2*np.pi*n*frequency*t))
    if y_sum is None:
        y_sum = yn
    else:
        y_sum += yn
    if n in (1, 3, 5):
        ax1.plot(t, y_sum)
ax1.plot(t, y_sum)
ax1.set_xlim(0, 2/frequency)
ax1.set_ylim(-1.5, 1.5)

yf = fft(y)
tf = np.linspace(.0, 1./(2.*dt), N//2)
spectrum = 2./N * np.abs(yf[0:N//2])

ax2.plot(tf, spectrum)
n = np.arange(1, 20, 2)
ax2.plot(n*frequency, 4/(np.pi*n), 'o', color='red')
ax2.grid()
ax2.set_title('Spectrum')
ax2.set_xlabel('Frequency [Hz]')
ax2.set_ylabel('Amplitude')

#f# save_figure('figure2', 'fft-square-waveform.png')

####################################################################################################

plt.show()



####################################################################################################

import numpy as np

import matplotlib.pyplot as plt

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Plot.BodeDiagram import bode_diagram
from PySpice.Probe.Plot import plot
from PySpice.Spice.Netlist import Circuit
from PySpice.Unit import *

from OperationalAmplifier import BasicOperationalAmplifier

#f# literal_include('OperationalAmplifier.py')

####################################################################################################

circuit = Circuit('Operational Amplifier')

# AC 1 PWL(0US 0V  0.01US 1V)
circuit.SinusoidalVoltageSource('input', 'in', circuit.gnd, amplitude=1@u_V)
circuit.subcircuit(BasicOperationalAmplifier())
circuit.X('op', 'BasicOperationalAmplifier', 'in', circuit.gnd, 'out')
circuit.R('load', 'out', circuit.gnd, 470@u_Ω)

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.ac(start_frequency=1@u_Hz, stop_frequency=100@u_MHz, number_of_points=5,  variation='dec')

figure, (ax1, ax2) = plt.subplots(2, figsize=(20, 10))

plt.title("Bode Diagram of an Operational Amplifier")
bode_diagram(axes=(ax1, ax2),
             frequency=analysis.frequency,
             gain=20*np.log10(np.absolute(analysis.out)),
             phase=np.angle(analysis.out, deg=False),
             marker='.',
             color='blue',
             linestyle='-',
            )
plt.tight_layout()
plt.show()

#f# save_figure('figure', 'operational-amplifier.png')



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

import unittest

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

from PySpice.Spice.Netlist import *
from PySpice.Unit import *

####################################################################################################

class VoltageDivider(SubCircuitFactory):
    NAME = 'VoltageDivider'
    NODES = ('input', 'output_plus', 'output_minus')
    def __init__(self):
        super().__init__()
        self.R(1, 'input', 'output_plus', 9@u_kΩ)
        self.R(2, 'output_plus', 'output_minus', 1@u_kΩ)

####################################################################################################

class VoltageDividerCircuit(Circuit):
    def __init__(self):
        super().__init__(title='Voltage Divider')
        self.V('input', 'in', self.gnd, '10V')
        self.R(1, 'in', 'out', 9@u_kΩ)
        self.R(2, 'out', self.gnd, 1@u_kΩ)

####################################################################################################

class TestNetlist(unittest.TestCase):

    ##############################################

    def _test_spice_declaration(self, circuit, spice_declaration):
        # Ignore line endings and blank lines
        circuit_lines = [*filter(None, str(circuit).splitlines(False))]
        expected_lines = [*filter(None, spice_declaration.splitlines(False))]
        self.assertEqual(circuit_lines, expected_lines)

    ##############################################

    def _test_nodes(self, circuit, nodes):

        node_names = sorted([str(node.name) for node in circuit.nodes])
        nodes = sorted([str(node) for node in nodes])
        self.assertListEqual(node_names, nodes)

####################################################################################################

class TestSubCircuit(TestNetlist):

    ##############################################

    def test(self):

        spice_declaration = """
.subckt VoltageDivider input output_plus output_minus
R1 input output_plus 9kOhm
R2 output_plus output_minus 1kOhm
.ends VoltageDivider
"""
        self._test_spice_declaration(VoltageDivider(), spice_declaration)

####################################################################################################

class TestCircuit(TestNetlist):

    ##############################################

    def test_basic(self):

        spice_declaration = """
.title Voltage Divider
Vinput in 0 10V
R1 in out 9kOhm
R2 out 0 1kOhm
"""
# .end

        circuit = Circuit('Voltage Divider')
        circuit.V('input', 'in', circuit.gnd, '10V')
        circuit.R(1, 'in', 'out', 9@u_kΩ)
        circuit.R(2, circuit.out, circuit.gnd, 1@u_kΩ) # out node is defined
        self._test_spice_declaration(circuit, spice_declaration)

        circuit = VoltageDividerCircuit()
        self._test_spice_declaration(circuit, spice_declaration)

        self._test_nodes(circuit, (0, 'in', 'out'))

        self.assertTrue(circuit.R1.minus.node is circuit.out)

        self.assertEqual(str(circuit.R1.plus.node), 'in')
        self.assertEqual(str(circuit.R1.minus.node), 'out')

        self.assertEqual(str(circuit['in']), 'in')
        self.assertEqual(str(circuit['out']), 'out')
        self.assertEqual(str(circuit.out), 'out')

        # for pin in circuit.out:
        #     print(pin)

        self.assertEqual(circuit.out.pins, set((circuit.R1.minus, circuit.R2.plus)))

        self.assertEqual(circuit.R1.resistance, 9@u_kΩ)
        self.assertEqual(circuit['R2'].resistance, 1@u_kΩ)

        circuit.R1.resistance = 10@u_kΩ
        self._test_spice_declaration(circuit, spice_declaration.replace('9k', '10k'))

        # .global .param .include .model

    ##############################################

    def test_ground_node(self):

        circuit = Circuit('')
        circuit.V('input', 'in', circuit.gnd, '10V')
        circuit.R(1, 'in', 'out', 9@u_kΩ)
        circuit.R(2, 'out', circuit.gnd, 1@u_kΩ)

        self.assertTrue(circuit.has_ground_node())

        circuit = Circuit('')
        circuit.V('input', 'in', 'fake_ground', '10V')
        circuit.R(1, 'in', 'out', 9@u_kΩ)
        circuit.R(2, 'out', 'fake_ground', 1@u_kΩ)

        self.assertFalse(circuit.has_ground_node())

    ##############################################

    def test_raw_spice(self):

        spice_declaration = """
.title Voltage Divider
R2 out 0 1kOhm
Vinput in 0 10V
R1 in out 9kOhm
"""
# .end

        circuit = Circuit('Voltage Divider')
        circuit.V('input', 'in', circuit.gnd, '10V')
        circuit.R(1, 'in', 'out', raw_spice='9kOhm')
        circuit.raw_spice += 'R2 out 0 1kOhm'
        self._test_spice_declaration(circuit, spice_declaration)

    ##############################################

    def test_keyword_clash(self):

        circuit = Circuit('')
        model = circuit.model('Diode', 'D', is_=1, rs=2)
        self.assertEqual(model.is_, 1)
        self.assertEqual(model['is'], 1)
        self.assertEqual(str(model), '.model Diode D (is=1 rs=2)')

    ##############################################

    def test_param(self):

        spice_declaration = """
.title Parameter Test
.param pippo=5
.param po=6
.param pp=7.8
.param pap={AGAUSS(pippo, 1 , 1.67)}
.param pippp={pippo + pp}
.param p={pp}
"""
# .param pop='pp +p'
# .end

        circuit = Circuit('Parameter Test')
        circuit.parameter('pippo', 5)
        circuit.parameter('po', 6)
        circuit.parameter('pp', 7.8)
        circuit.parameter('pap', '{AGAUSS(pippo, 1 , 1.67)}')
        circuit.parameter('pippp', '{pippo + pp}')
        circuit.parameter('p', '{pp}')
        # circuit.parameter('pop', 'pp + p')
        self._test_spice_declaration(circuit, spice_declaration)

####################################################################################################

if __name__ == '__main__':

    unittest.main()



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

# Used by Spice/Library.py

####################################################################################################

import os
import subprocess

####################################################################################################

def file_name_has_extension(file_name, extension):

    return file_name.endswith(extension)

####################################################################################################

def file_extension(filename):

    # index = filename.rfind(os.path.extsep)
    # if index == -1:
    #     return None
    # else:
    #     return filename[index:]

    return os.path.splitext(filename)[1]

####################################################################################################

def run_shasum(filename, algorithm=1, text=False, binary=False, portable=False):

    if algorithm not in (1, 224, 256, 384, 512, 512224, 512256):
        raise ValueError

    args = ['shasum', '--algorithm=' + str(algorithm)]
    if text:
        args.append('--text')
    elif binary:
        args.append('--binary')
    elif portable:
        args.append('--portable')
    args.append(filename)
    output = subprocess.check_output(args)
    shasum = output[:output.find(' ')]

    return shasum

####################################################################################################

class Path:

    ##############################################

    def __init__(self, path):

        self._path = str(path)

    ##############################################

    def __bool__(self):

        return os.path.exists(self._path)

    ##############################################

    def __str__(self):

        return self._path

    ##############################################

    @property
    def path(self):

        return self._path

    ##############################################

    def is_absolut(self):

        return os.path.isabs(self._path)

    ##############################################

    def absolut(self):

        return self.clone_for_path(os.path.abspath(self._path))

    ##############################################

    def normalise(self):

        return self.clone_for_path(os.path.normpath(self._path))

    ##############################################

    def normalise_case(self):

        return self.clone_for_path(os.path.normcase(self._path))

    ##############################################

    def expand_vars_and_user(self):

        return self.clone_for_path(os.path.expandvars(os.path.expanduser(self._path)))

    ##############################################

    def real_path(self):

        return self.clone_for_path(os.path.realpath(self._path))

    ##############################################

    def relative_to(self, directory):

        return self.clone_for_path(os.path.relpath(self._path, str(directory)))

    ##############################################

    def clone_for_path(self, path):

        return self.__class__(path)

    ##############################################

    def split(self):

        return self._path.split(os.path.sep)

    ##############################################

    def directory_part(self):

        return Directory(os.path.dirname(self._path))

    ##############################################

    def filename_part(self):

        return os.path.basename(self._path)

    ##############################################

    def is_directory(self):

        return os.path.isdir(self._path)

    ##############################################

    def is_file(self):

        return os.path.isfile(self._path)

    ##############################################

    @property
    def inode(self):

        return os.stat(self._path).st_ino

    ##############################################

    @property
    def creation_time(self):

        return os.stat(self._path).st_ctime

####################################################################################################

class Directory(Path):

    ##############################################

    def __bool__(self):

        return super().__nonzero__() and self.is_directory()

    ##############################################

    def join_directory(self, directory):

        return self.__class__(os.path.join(self._path, str(directory)))

    ##############################################

    def join_filename(self, filename):

        return File(filename, self._path)

    ##############################################

    def iter_file(self, followlinks=False):

        if self.is_file():
            yield File(self.filename_part(), self.directory_part())
        else:
            for root, directories, files in os.walk(self._path, followlinks=followlinks):
                for filename in files:
                    yield File(filename, root)

    ##############################################

    def iter_directories(self, followlinks=False):

        for root, directories, files in os.walk(self._path, followlinks=followlinks):
            for directory in directories:
                yield Path(os.path.join(root, directory))

####################################################################################################

class File(Path):

    default_shasum_algorithm = 256

    ##############################################

    def __init__(self, filename, path=''):

        super().__init__(os.path.join(str(path), str(filename)))

        self._filename = self.filename_part()
        if not self._filename:
            raise ValueError
        self._directory = self.directory_part()

        self._shasum = None # lazy computation

    ##############################################

    def __bool__(self):

        return super().__nonzero__() and os.path.isfile(self._path)

    ##############################################

    @property
    def directory(self):

        return self._directory

    ##############################################

    @property
    def filename(self):

        return self._filename

    ##############################################

    @property
    def extension(self):

        return file_extension(self._filename)

    ##############################################

    @property
    def shasum(self):

        if self._shasum is None:
            return self.compute_shasum()
        else:
            return self._shasum

    ##############################################

    def compute_shasum(self, algorithm=None):

        if algorithm is None:
            algorithm = self.default_shasum_algorithm
        self._shasum = run_shasum(self._path, algorithm, portable=True)

        return self._shasum



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

"""This module provides algorithms to compute the derivative of a function sampled on an uniform
grid.
"""

####################################################################################################

import fractions

import numpy as np

####################################################################################################

from PySpice.Math import odd

####################################################################################################

def compute_exact_finite_difference_coefficients(derivative_order, grid, x0=0):

    """This function compute the finite difference coefficients for the given derivative order and
    grid.  The parameter *x* specifies where is computed the derivative on the grid.  The grid is
    given as a list of integer offsets.

    The algorithm is derived from the article: Generation of Finite Difference Formulas on Arbitrary Space
    Grids, Bengt Fornberg, Mathematics of computation, volume 51, number 184, october 1988
    """

    N = len(grid)

    # d[m,n,v]
    d = [[[0
           for v in range(N)]
          for n in range(N)]
         for m in range(derivative_order +1)]

    d[0][0][0] = fractions.Fraction(1,1)
    c1 = 1
    for n in range(1, N):
        c2 = 1
        for v in range(n):
            c3 = grid[n] - grid[v]
            c2 *= c3
            if n <= derivative_order:
                d[n][n-1][v] = 0
            for m in range(min(n, derivative_order) +1):
                d[m][n][v] = ( (grid[n] - x0)*d[m][n-1][v] - m*d[m-1][n-1][v] ) / c3
        for m in range(min(n, derivative_order) +1):
            d[m][n][n] = fractions.Fraction(c1,c2)*( m*d[m-1][n-1][n-1] - (grid[n-1] - x0)*d[m][n-1][n-1] )
        c1 = c2

    return d[-1][-1]

####################################################################################################

def compute_finite_difference_coefficients(derivative_order, grid):
    return [float(x) for x in compute_exact_finite_difference_coefficients(derivative_order, grid)]

####################################################################################################

_coefficient_cache = dict(centred={}, forward={}, backward={})

def get_finite_difference_coefficients(derivative_order, accuracy_order, grid_type):

    if derivative_order < 1:
        raise ValueError("Wrong derivative order")

    if odd(accuracy_order) or accuracy_order < 2:
        raise ValueError("Wrong accuracy order")

    if grid_type == 'centred':
        window_size = accuracy_order // 2
        grid = list(range(-window_size, window_size +1))
    elif grid_type == 'forward':
        grid = list(range(derivative_order + accuracy_order))
    elif grid_type == 'backward':
        grid = list(range(-(derivative_order + accuracy_order) +1, 1))
        grid = list(reversed(grid)) # Fixme: why ?
    else:
        raise ValueError("Wrong grid type")

    key = '{}-{}'.format(derivative_order, accuracy_order)
    coefficients = _coefficient_cache[grid_type].get(key, None)
    if coefficients is None:
        coefficients = compute_finite_difference_coefficients(derivative_order, grid)
        _coefficient_cache[grid_type][key] = coefficients

    return grid, coefficients

####################################################################################################

def simple_derivative(x, values):
    """ Compute the derivative as a simple slope. """
    return x[:-1], np.diff(values)/np.diff(x)

####################################################################################################

def derivative(x, values, derivative_order=1, accuracy_order=4):

    """Compute the derivative at the given derivative order and accuracy order. The precision of the
    Taylor expansion is :math:`\mathcal{O}(dx^{accuracy})`.
    """

    dx = np.diff(x)
    # if not np.all(dx == dx[0]):
    #     raise ValueError("Sampling is not uniform")
    dx = dx[0]

    values_size, = values.shape

    derivative = np.zeros(values_size, dtype=values.dtype)

    grid, coefficients = get_finite_difference_coefficients(derivative_order, accuracy_order, 'centred')
    window_size = grid[-1]
    # print grid, coefficients
    vector_size = values_size - 2*window_size
    if not vector_size:
        raise ValueError("The size of the value's array is not sufficient for the given accuracy order")
    lower_index = window_size
    upper_index = values_size - window_size
    derivative_view = derivative[window_size:-window_size]
    for offset, coefficient in zip(grid, coefficients):
        if coefficient:
            # print offset, lower_index + offset, upper_index + offset
            derivative_view += values[lower_index + offset:upper_index + offset] * coefficient

    grid, coefficients = get_finite_difference_coefficients(derivative_order, accuracy_order, 'forward')
    # print grid, coefficients
    grid_size = len(grid)
    upper_index = window_size
    derivative_view = derivative[:window_size]
    for offset, coefficient in zip(grid, coefficients):
        # print offset, offset, window_size+offset
        derivative_view += values[offset:upper_index + offset] * coefficient

    grid, coefficients = get_finite_difference_coefficients(derivative_order, accuracy_order, 'backward')
    # print grid, coefficients
    grid_size = len(grid)
    lower_index = values_size - window_size
    upper_index = values_size
    derivative_view = derivative[-window_size:]
    for offset, coefficient in zip(grid, coefficients):
        # print offset, lower_index + offset, upper_index + offset
        derivative_view += values[lower_index + offset:upper_index + offset] * coefficient

    return derivative / dx**derivative_order



#r# This example shows how to use units

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Spice.Netlist import Circuit

####################################################################################################

from PySpice.Unit import *

foo = kilo(1) # unit less

resistance_unit = U_Ω

resistance1 = u_kΩ(1)
resistance1 = u_kOhm(1) # ASCII variant

resistance1 = 1@u_kΩ   # using Python 3.5 syntax
resistance1 = 1 @u_kΩ  # space doesn't matter
resistance1 = 1 @ u_kΩ #

resistance2 = as_Ω(resistance1) # check unit

resistances = u_kΩ(range(1, 11)) # same as [u_kΩ(x) for x in range(1, 11)]
resistances = range(1, 11)@u_kΩ  # using Python 3.5 syntax

capacitance = u_uF(200)
inductance = u_mH(1)
temperature = u_Degree(25)

voltage = resistance1 * u_mA(1) # compute unit

frequency = u_ms(20).frequency
period = u_Hz(50).period
pulsation = frequency.pulsation
pulsation = period.pulsation

#r# According to the Python `operator precedence
#r# <https://docs.python.org/3/reference/expressions.html#operator-precedence>`_, division operators
#r# have a higher priority than the matrix multiplication operator.  In consequence you must had
#r# parenthesis to perform something like :code:`(10@u_s) / (2@_us)`.

####################################################################################################

circuit = Circuit('Resistor Bridge')

resistance = 10@u_kΩ
print(float(resistance))
print(str(resistance))

circuit.V('input', 1, circuit.gnd, 10@u_V)
circuit.R(1, 1, 2, 2@u_kΩ)
circuit.R(2, 1, 3, 1@u_kΩ)
circuit.R(3, 2, circuit.gnd, 1@u_kΩ)
circuit.R(4, 3, circuit.gnd, 2@u_kΩ)
circuit.R(5, 3, 2, 2@u_kΩ)

print(circuit)

####################################################################################################

import pint
u = pint.UnitRegistry()

resistance = 10*u.kΩ
# print(float(resistance))
print(resistance.magnitude)
print(resistance.m)
print(resistance.units)
print(str(resistance))

circuit = Circuit('Resistor Bridge')

circuit.V('input', 1, circuit.gnd, 10*u.V)
circuit.R(1, 1, 2, 2*u.kΩ)
circuit.R(2, 1, 3, 1*u.kΩ)
circuit.R(3, 2, circuit.gnd, 1*u.kΩ)
circuit.R(4, 3, circuit.gnd, 2*u.kΩ)
circuit.R(5, 3, 2, 2*u.kΩ)

print(circuit)



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

"""This module implements units.

A shortcut is defined for each unit prefix, e.g. :class:`pico`, :class:`nano`, :class:`micro`,
:class:`milli`, :class:`kilo`, :class:`mega`, :class:`tera`.

"""

# https://numpy.org/doc/stable/user/basics.subclassing.html#basics-subclassing

####################################################################################################

import logging

import collections.abc as collections
import math
# import numbers

import numpy as np

####################################################################################################

from PySpice.Tools.EnumFactory import EnumFactory

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class UnitPrefixMetaclass(type):

    """Metaclass to register unit prefixes"""

    _prefixes = {} # singletons

    ##############################################

    def __new__(meta, class_name, base_classes, attributes):
        cls = type.__new__(meta, class_name, base_classes, attributes)
        if class_name != 'UnitPrefix':
            meta.register_prefix(cls)
        return cls

    ##############################################

    @classmethod
    def register_prefix(meta, cls):
        power = cls.POWER
        if power is None:
            raise ValueError('Power is None for {}'.format(cls.__name__))
        meta._prefixes[power] = cls()

    ##############################################

    @classmethod
    def prefix_iter(cls):
        return cls._prefixes.values()

    ##############################################

    @classmethod
    def get(cls, power):
        return cls._prefixes[power]

####################################################################################################

class UnitPrefix(metaclass=UnitPrefixMetaclass):

    """This class implements a unit prefix like kilo"""

    POWER = None
    PREFIX = ''

    ##############################################

    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__, self.POWER, self.PREFIX)

    ##############################################

    def __int__(self):
        return self.POWER

    ##############################################

    def __str__(self):
        return self.PREFIX

    ##############################################

    @property
    def power(self):
        return self.POWER

    @property
    def prefix(self):
        return self.PREFIX

    @property
    def is_unit(self):
        return self.POWER == 0

    @property
    def scale(self):
        return 10**self.POWER

    ##############################################

    @property
    def spice_prefix(self):
        if hasattr(self, 'SPICE_PREFIX'):
            return self.SPICE_PREFIX
        else:
            return self.PREFIX

    ##############################################

    @property
    def is_defined_in_spice(self):
        return self.spice_prefix is not None

    ##############################################

    def __eq__(self, other):
        return self.POWER == other.POWER

    ##############################################

    def __ne__(self, other):
        return self.POWER != other.POWER

    ##############################################

    def __lt__(self, other):
        return self.POWER < other.POWER

    ##############################################

    def __gt__(self, other):
        return self.POWER > other.POWER

    ##############################################

    def str(self, spice=False):
        if spice:
            return self.spice_prefix
        else:
            return self.PREFIX

####################################################################################################

class ZeroPower(UnitPrefix):
    POWER = 0
    PREFIX = ''
    SPICE_PREFIX = ''

_zero_power = UnitPrefixMetaclass.get(0)

####################################################################################################

class SiDerivedUnit:

    """This class implements a unit defined as powers of SI base units.
    """

    # SI base units
    BASE_UNITS = (
        'm',
        'kg',
        's',
        'A',
        'K',
        'mol',
        'cd',
    )

    ##############################################

    def __init__(self, string=None, powers=None):

        if powers is not None:
            self._powers = self.new_powers()
            self._powers.update(powers)
        elif string is not None:
            self._powers = self.parse_si(string)
        else:
            self._powers = self.new_powers()

        self._hash = self.to_hash(self._powers)
        self._string = self.to_string(self._powers)

    ##############################################

    @property
    def powers(self):
        return self._powers

    @property
    def hash(self):
        return self._hash

    @property
    def string(self):
        return self._string

    def __str__(self):
        return self._string

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self._string)

    ##############################################

    @classmethod
    def new_powers(cls):
        return {unit: 0 for unit in cls.BASE_UNITS}

    ##############################################

    @classmethod
    def parse_si(cls, string):
        si_powers = cls.new_powers()
        if string:
            for prefixed_units in string.split('*'):
                parts = prefixed_units.split('^')
                unit = parts[0]
                if len(parts) == 1:
                    powers = 1
                else:
                    powers = int(parts[1])
                si_powers[unit] += powers
        return si_powers

    ##############################################

    @classmethod
    def to_hash(cls, powers):
        hash_ = ''
        for unit in cls.BASE_UNITS:
            hash_ += str(powers[unit])
        return hash_

    ##############################################

    @classmethod
    def to_string(cls, si_powers):
        units = []
        for unit in cls.BASE_UNITS:
            powers = si_powers[unit]
            if powers == 1:
                units.append(unit)
            elif powers > 1 or powers < 0:
                units.append('{}^{}'.format(unit, powers))
        return '*'.join(units)

    ##############################################

    # @property
    def is_base_unit(self):
        count = 0
        for powers in self._powers.values():
            if powers == 1:
                count += 1
            elif powers != 0:
                return False
        return count == 1

    ##############################################

    # @property
    def is_unit_less(self):
        return self._hash == '0'*len(self.BASE_UNITS)

    ##############################################

    def __bool__(self):
        return not self.is_unit_less()

    ##############################################

    def clone(self):
        return self.__class__(powers=self._powers)

    ##############################################

    def __eq__(self, other):
        return self._hash == other.hash

    ##############################################

    def __ne__(self, other):
        return self._hash != other.hash

    ##############################################

    def __mul__(self, other):
        powers = {unit: self._powers[unit] + other._powers[unit]
                  for unit in self.BASE_UNITS}
        return self.__class__(powers=powers)

    ##############################################

    def __imul__(self, other):
        for unit in self.BASE_UNITS:
            self._powers[unit] += other.powers[unit]
        self._hash = self.to_hash(self._powers)
        self._string = self.to_string(self._powers)
        return self

    ##############################################

    def __truediv__(self, other):
        powers = {unit: self._powers[unit] - other._powers[unit]
                  for unit in self.BASE_UNITS}
        return self.__class__(powers=powers)

    ##############################################

    def __itruediv__(self, other):
        for unit in self.BASE_UNITS:
            self._powers[unit] -= other.powers[unit]
        self._hash = self.to_hash(self._powers)
        self._string = self.to_string(self._powers)
        return self

    ##############################################

    def power(self, value):
        powers = {unit: self._powers[unit] * value
                  for unit in self.BASE_UNITS}
        return self.__class__(powers=powers)

    ##############################################

    def reciprocal(self):
        return self.power(-1)

    ##############################################

    def sqrt(self):
        return self.power(1/2)

    ##############################################

    def square(self):
        return self.power(2)

    ##############################################

    def cbrt(self):
        return self.power(1/3)

####################################################################################################

class UnitMetaclass(type):

    """Metaclass to register units"""

    _units = {}
    _hash_map = {}

    ##############################################

    def __new__(meta, class_name, base_classes, attributes):
        cls = type.__new__(meta, class_name, base_classes, attributes)
        meta.init_unit(cls)
        meta.register_unit(cls)
        return cls

    ##############################################

    @classmethod
    def init_unit(meta, cls):

        si_unit = cls.SI_UNIT
        if not (isinstance(si_unit, SiDerivedUnit) and si_unit):
            # si_unit is not defined
            if cls.is_base_unit():
                si_unit = SiDerivedUnit(cls.UNIT_SUFFIX)
            else: # str
                si_unit = SiDerivedUnit(si_unit)
            cls.SI_UNIT = si_unit

    ##############################################

    @classmethod
    def register_unit(meta, cls):

        obj = cls()
        meta._units[obj.unit_suffix] = obj

        if obj.si_unit:
            hash_ = obj.si_unit.hash
            if hash_ in meta._hash_map:
                meta._hash_map[hash_].append(obj)
            else:
                meta._hash_map[hash_] = [obj]

    ##############################################

    @classmethod
    def unit_iter(meta):
        return meta._units.values()

    ##############################################

    @classmethod
    def from_prefix(meta, prefix):
        return meta._units__.get(prefix, None)

    ##############################################

    @classmethod
    def from_hash(meta, hash_):
        return meta._hash_map.get(hash_, None)

    ##############################################

    @classmethod
    def from_si_unit(meta, si_unit, unique=True):

        # Fixme:
        #  - handle power of units
        #      unit -> numpy vector, divide and test for identical factor
        #      define unit, format as V^2
        #  - complex unit

        units = meta._hash_map.get(si_unit.hash, None)
        if unique and units is not None:
            if len(units) > 1:
                units = [unit for unit in units if unit.is_default_unit()]
                if len(units) == 1:
                    return units[0]
                else:
                    raise NameError("Unit clash", units)
            else:
                return units[0]
        else:
            return units

####################################################################################################

class UnitError(ValueError):
    pass

####################################################################################################

class Unit(metaclass=UnitMetaclass):

    """This class implements a unit.
    """

    UNIT_NAME = ''
    UNIT_SUFFIX = ''
    QUANTITY = ''
    SI_UNIT = SiDerivedUnit()
    DEFAULT_UNIT = False
    # SPICE_SUFFIX = ''

    _logger = _module_logger.getChild('Unit')

    ##############################################

    def __init__(self, si_unit=None):

        self._unit_name = self.UNIT_NAME
        self._unit_suffix = self.UNIT_SUFFIX
        self._quantity = self.QUANTITY

        if si_unit is None:
            self._si_unit = self.SI_UNIT
        else:
            self._si_unit = si_unit

    ##############################################

    def __repr__(self):
        return '{0}({1})'.format(self.__class__.__name__, str(self))

    ##############################################

    @property
    def unit_name(self):
        return self._unit_name

    @property
    def unit_suffix(self):
        return self._unit_suffix

    @property
    def quantity(self):
        return self._quantity

    @property
    def si_unit(self):
        return self._si_unit

    ##############################################

    @property
    def is_unit_less(self):
        return self._si_unit.is_unit_less()

    ##############################################

    @classmethod
    def is_default_unit(cls):
        return cls.DEFAULT_UNIT

    @classmethod
    def is_base_unit(cls):
        return False

    ##############################################

    def __eq__(self, other):
        """self == other"""
        return self._si_unit == other.si_unit

    ##############################################

    def __ne__(self, other):
        """self != other"""
        # The default __ne__ doesn't negate __eq__ until 3.0.
        return not (self == other)

    ##############################################

    def _equivalent_prefixed_unit(self, si_unit):

        equivalent_unit = PrefixedUnit.from_si_unit(si_unit)
        if equivalent_unit is not None:
            return equivalent_unit
        else:
            return PrefixedUnit(Unit(si_unit))

    ##############################################

    def _equivalent_unit(self, si_unit):

        equivalent_unit = UnitMetaclass.from_si_unit(si_unit)
        if equivalent_unit is not None:
            return equivalent_unit
        else:
            return Unit(si_unit)

    ##############################################

    def _equivalent_unit_or_power(self, si_unit, prefixed_unit):
        if prefixed_unit:
            return self._equivalent_prefixed_unit(si_unit)
        else:
            return self._equivalent_unit(si_unit)

    ##############################################

    def multiply(self, other, prefixed_unit=False):
        si_unit = self._si_unit * other.si_unit
        return self._equivalent_unit_or_power(si_unit, prefixed_unit)

    ##############################################

    def divide(self, other, prefixed_unit=False):
        si_unit = self._si_unit / other.si_unit
        return self._equivalent_unit_or_power(si_unit, prefixed_unit)

    ##############################################

    def power(self, exponent, prefixed_unit=False):
        si_unit = self._si_unit.power(exponent)
        return self._equivalent_unit_or_power(si_unit, prefixed_unit)

    ##############################################

    def reciprocal(self, prefixed_unit=False):
        si_unit = self._si_unit.reciprocal()
        return self._equivalent_unit_or_power(si_unit, prefixed_unit)

    ##############################################

    def sqrt(self, prefixed_unit=False):
        si_unit = self._si_unit.sqrt()
        return self._equivalent_unit_or_power(si_unit, prefixed_unit)

    ##############################################

    def square(self, prefixed_unit=False):
        si_unit = self._si_unit.square()
        return self._equivalent_unit_or_power(si_unit, prefixed_unit)

    ##############################################

    def cbrt(self, prefixed_unit=False):
        si_unit = self._si_unit.cbrt()
        return self._equivalent_unit_or_power(si_unit, prefixed_unit)

    ##############################################

    def __str__(self):
        if self._unit_suffix:
            return self._unit_suffix
        else:
            return str(self._si_unit)

    ##############################################

    def is_same_unit(self, value):
        return value.unit == self

    ##############################################

    def validate(self, value, none=False):

        if none and value is None:
            return None
        if isinstance(value, UnitValue):
            if  self.is_same_unit(value):
                return value
            else:
                raise UnitError
        else:
            prefixed_unit = PrefixedUnit.from_prefixed_unit(self)
            return prefixed_unit.new_value(value)

####################################################################################################

class SiBaseUnit(Unit):

    """This class implements an SI base unit."""

    ##############################################

    @classmethod
    def is_base_unit(cls):
        return True

    ##############################################

    @classmethod
    def is_default_unit(cls):
        return True

####################################################################################################

class PrefixedUnit:

    """This class implements a prefixed unit.
    """

    _unit_map = {} # Prefixed unit singletons
    _prefixed_unit_map = {}

    _value_ctor = None
    _values_ctor = None

    ##############################################

    @classmethod
    def register(cls, prefixed_unit):
        unit = prefixed_unit.unit
        unit_prefix = prefixed_unit.power
        if unit_prefix.is_unit and unit.is_default_unit():
            key = unit.si_unit.hash
            # print('Register', key, prefixed_unit)
            cls._unit_map[key] = prefixed_unit
        if unit.unit_suffix:
            unit_key = str(unit)
        else:
            unit_key = '_'
        power_key = unit_prefix.power
        # print('Register', unit_key, power_key, prefixed_unit)
        if unit_key not in cls._prefixed_unit_map:
            cls._prefixed_unit_map[unit_key] = {}
        cls._prefixed_unit_map[unit_key][power_key] = prefixed_unit

    ##############################################

    @classmethod
    def from_si_unit(cls, si_unit):
        return cls._unit_map.get(si_unit.hash, None)

    ##############################################

    @classmethod
    def from_prefixed_unit(cls, unit, power=0):

        if unit.unit_suffix:
            unit_key = str(unit)
        else:
            if power == 0:
                return _simple_prefixed_unit
            unit_key = '_'
        try:
            return cls._prefixed_unit_map[unit_key][power]
        except KeyError:
            return None

    ##############################################

    def __init__(self, unit=None, power=None, value_ctor=None, values_ctor=None):

        if unit is None:
            self._unit = Unit()
        else:
            self._unit = unit
        if power is None:
            self._power = _zero_power
        else:
            self._power = power

        if value_ctor is not None:
            self._value_ctor = value_ctor

        if values_ctor is not None:
            self._values_ctor = values_ctor

    ##############################################

    def __repr__(self):
        return '{0}({1})'.format(self.__class__.__name__, str(self))

    ##############################################

    @property
    def unit(self):
        return self._unit

    @property
    def power(self):
        return self._power

    @property
    def scale(self):
        return self._power.scale

    ##############################################

    @property
    def is_unit_less(self):
        return self._unit.is_unit_less

    ##############################################

    def clone(self):
        return self.__class__(self._unit, self._power)

    ##############################################

    def is_same_unit(self, other):
        return self._unit == other.unit

    ##############################################

    def check_unit(self, other):
        if not self.is_same_unit(other):
            raise UnitError('{} versus {}'.format(self, other))

    ##############################################

    def is_same_power(self, other):
        return self._power == other.power

    ##############################################

    def __eq__(self, other):
        """self == other"""
        return self.is_same_unit(other) and self.is_same_power(other)

    ##############################################

    def __ne__(self, other):
        """self != other"""
        # The default __ne__ doesn't negate __eq__ until 3.0.
        return not (self == other)

    ##############################################

    def str(self, spice=False, unit=True):

        # Ngspice User Manual Section 2.3.1  Some naming conventions
        #
        # Letters immediately following a number that are not scale factors are ignored, and
        # letters immediately following a scale factor are ignored.
        #
        # Hence, 10, 10V, 10Volts, and 10Hz all represent the same number, and
        # M, MA, MSec, and  MMhos all represent the same scale factor.
        #
        # Note that 1000, 1000.0, 1000Hz, 1e3, 1.0e3, 1kHz, and 1k all represent the same number.

        # >>> WARNING <<<
        #   Note that M or m denote ’milli’, i.e. 10−3 . Suffix meg has to be used for 106.
        #   see SPICE_PREFIX in SiUnits

        # Fixme: unit clash, e.g. mm ???

        string = self._power.str(spice)

        if unit:
            string += str(self._unit)

        if spice:
            # F is interpreted as f = femto
            if string == 'F':
                string = ''
            else:
                # Ngspice don't support utf-8
                # degree symbole can be encoded str(176) in Extended ASCII
                string = string.replace('°', '')  # U+00B0
                string = string.replace('℃', '')  # U+2103
                # U+2109 ℉
                string = string.replace('Ω', 'Ohm')  # U+CEA0
                string = string.replace('μ',   'u')  # U+CEBC

        return string

    ##############################################

    def str_spice(self):
        return self.str(spice=True, unit=True)

    ##############################################

    def __str__(self):
        return self.str(spice=False, unit=True)

    ##############################################

    def new_value(self, value):
        if isinstance(value, np.ndarray):
            return self._values_ctor.from_ndarray(value, self)
        elif isinstance(value, collections.Iterable):
            return [self._value_ctor(self, x) for x in value]
        else:
            return self._value_ctor(self, value)

####################################################################################################

class UnitValue: # numbers.Real

    """This class implements a value with a unit and a power (prefix).

    The value is not converted to float if the value is an int.
    """

    _logger = _module_logger.getChild('UnitValue')

    ##############################################

    @classmethod
    def simple_value(cls, value):
        return cls(_simple_prefixed_unit, value)

    ##############################################

    def __init__(self, prefixed_unit, value):

        self._prefixed_unit = prefixed_unit

        if isinstance(value, UnitValue):
            # Fixme: anonymous ???
            if not self.is_same_unit(value):
                raise UnitError
            if self.is_same_power(value):
                self._value = value.value
            else:
                self._value = self._convert_scalar_value(value)
        elif isinstance(value, int):
            self._value = value # to keep as int
        else:
            self._value = float(value)

    ##############################################

    def __repr__(self):
        return '{0}({1})'.format(self.__class__.__name__, str(self))

    ##############################################

    @property
    def prefixed_unit(self):
        return self._prefixed_unit

    @property
    def unit(self):
        return self._prefixed_unit.unit

    @property
    def power(self):
        return self._prefixed_unit.power

    @property
    def scale(self):
        return self._prefixed_unit.power.scale

    @property
    def value(self):
        return self._value

    ##############################################

    def clone(self):
        return self.__class__(self._prefixed_unit, self._value)

    ##############################################

    def clone_prefixed_unit(self, value):
        return self.__class__(self._prefixed_unit, value)

    ##############################################

    # def to_unit_values(self):
    #     return self._prefixed_unit.new_value(self._value)

    ##############################################

    # def clone_unit(self, value, power):
    #     return self.__class__(PrefixedUnit(self.unit, power), value)

    ##############################################

    def is_same_unit(self, other):
        return self._prefixed_unit.is_same_unit(other.prefixed_unit)

    ##############################################

    def _check_unit(self, other):
        if not self.is_same_unit(other):
            raise UnitError

    ##############################################

    def is_same_power(self, other):
        return self._prefixed_unit.is_same_power(other.prefixed_unit)

    ##############################################

    def __eq__(self, other):
        """self == other"""
        if isinstance(other, UnitValue):
            return self.is_same_unit(other) and float(self) == float(other)
        else:
            return float(self) == float(other)

    ##############################################

    def __ne__(self, other):
        """self != other"""
        # The default __ne__ doesn't negate __eq__ until 3.0.
        return not (self == other)

    ##############################################

    def _convert_value(self, other):
        """Convert the value of other to the power of self."""
        self._check_unit(other)
        if self.is_same_power(other):
            return other.value
        else:
            return other.value * (other.scale / self.scale) # for numerical precision

    ##############################################

    def _convert_scalar_value(self, value):
        return float(value) / self.scale

    ##############################################

    def __int__(self):
        return int(self._value * self.scale)

    ##############################################

    def __float__(self):
        return float(self._value * self.scale)

    ##############################################

    def str(self, spice=False, space=False, unit=True):
        string = str(self._value)
        if space:
            string += ' '
        string += self._prefixed_unit.str(spice, unit)
        return string

    ##############################################

    def str_space(self):
        return self.str(space=True)

    ##############################################

    def str_spice(self):
        return self.str(spice=True, space=False, unit=True)

    ##############################################

    def __str__(self):
        return self.str(spice=False, space=True, unit=True)

    ##############################################

    def __bool__(self):
        """True if self != 0. Called for bool(self)."""
        return self._value != 0

    ##############################################

    def __add__(self, other):
        """self + other"""
        if (isinstance(other, UnitValue)):
            self._check_unit(other)
            new_obj = self.clone()
            new_obj._value += self._convert_value(other)
            return new_obj
        else:
            return float(self) + other

    ##############################################

    def __iadd__(self, other):
        """self += other"""
        self._check_unit(other)
        self._value += self._convert_value(other)
        return self

    ##############################################

    def __radd__(self, other):
        """other + self"""
        return float(self) + other

    ##############################################

    def __neg__(self):
        """-self"""
        return self.clone_prefixed_unit(-self._value)

    ##############################################

    def __pos__(self):
        """+self"""
        return self.clone()

    ##############################################

    def __sub__(self, other):
        """self - other"""
        if (isinstance(other, UnitValue)):
            self._check_unit(other)
            new_obj = self.clone()
            new_obj._value -= self._convert_value(other)
            return new_obj
        else:
            return float(self) - other

    ##############################################

    def __isub__(self, other):
        """self -= other"""
        self._check_unit(other)
        self._value -= self._convert_value(other)
        return self

    ##############################################

    def __rsub__(self, other):
        """other - self"""
        return other - float(self)

    ##############################################

    def __mul__(self, other):
        """self * other"""
        if (isinstance(other, UnitValue)):
            equivalent_unit = self.unit.multiply(other.unit, True)
            value = float(self) * float(other)
            return equivalent_unit.new_value(value)
        else:
            try: # scale value
                scalar = float(other)
                new_obj = self.clone()
                new_obj._value *= scalar
                return new_obj
            except (ValueError, TypeError): # Numpy raises TypeError
                return float(self) * other

    ##############################################

    def __imul__(self, other):
        """self *= other"""
        if (isinstance(other, UnitValue)):
            raise UnitError
        else: # scale value
            # Fixme: right ?
            self._value *= self._convert_value(other)
            return self

    ##############################################

    def __rmul__(self, other):
        """other * self"""
        if (isinstance(other, UnitValue)):
            raise NotImplementedError # Fixme: when ???
        else: # scale value
            return self.__mul__(other)

    ##############################################

    def __floordiv__(self, other):

        """self // other """

        if (isinstance(other, UnitValue)):
            equivalent_unit = self.unit.divide(other.unit, True)
            value = float(self) // float(other)
            return equivalent_unit.new_value(value)
        else:
            try: # scale value
                scalar = float(other)
                new_obj = self.clone()
                new_obj._value //= scalar
                return new_obj
            except (ValueError, TypeError): # Numpy raises TypeError
                return float(self) // other

    ##############################################

    def __ifloordiv__(self, other):
        """self //= other """
        if (isinstance(other, UnitValue)):
            raise NotImplementedError
        else: # scale value
            self._value //= float(other)
            return self

    ##############################################

    def __rfloordiv__(self, other):
        """other // self"""
        if (isinstance(other, UnitValue)):
            raise NotImplementedError # Fixme: when ???
        else: # scale value
            return other // float(self)

    ##############################################

    def __truediv__(self, other):

        """self / other"""

        if (isinstance(other, UnitValue)):
            equivalent_unit = self.unit.divide(other.unit, True)
            value = float(self) / float(other)
            return equivalent_unit.new_value(value)
        else:
            try: # scale value
                scalar = float(other)
                new_obj = self.clone()
                new_obj._value /= scalar
                return new_obj
            except (ValueError, TypeError): # Numpy raises TypeError
                return float(self) / other

    ##############################################

    def __itruediv__(self, other):
        """self /= other"""
        if (isinstance(other, UnitValue)):
            raise NotImplementedError
        else: # scale value
            self._value /= float(other)
            return self

    ##############################################

    def __rtruediv__(self, other):
        """other / self"""
        if (isinstance(other, UnitValue)):
            raise NotImplementedError # Fixme: when ???
        else: # scale value
            return other / float(self)

    ##############################################

    def __pow__(self, exponent):
        """self**exponent; should promote to float or complex when necessary."""
        new_obj = self.clone()
        new_obj._value **= float(exponent)
        return new_obj

    ##############################################

    def __ipow__(self, exponent):
        self._value **= float(exponent)
        return self

    ##############################################

    def __rpow__(self, base):
        """base ** self"""
        raise NotImplementedError

    ##############################################

    def __abs__(self):
        """Returns the Real distance from 0. Called for abs(self)."""
        return self.clone_prefixed_unit(abs(self._value))

    ##############################################

    def __trunc__(self):
        """trunc(self): Truncates self to an Integral.

        Returns an Integral i such that:
          * i>0 iff self>0;
          * abs(i) <= abs(self);
          * for any Integral j satisfying the first two conditions,
            abs(i) >= abs(j) [i.e. i has "maximal" abs among those].
        i.e. "truncate towards 0".
        """
        raise NotImplementedError

    ##############################################

    def __divmod__(self, other):
        """divmod(self, other): The pair (self // other, self % other).

        Sometimes this can be computed faster than the pair of
        operations.
        """
        return (self // other, self % other)

    ##############################################

    def __rdivmod__(self, other):
        """divmod(other, self): The pair (self // other, self % other).

        Sometimes this can be computed faster than the pair of
        operations.
        """
        return (other // self, other % self)

    ##############################################

    def __mod__(self, other):
        """self % other"""
        raise NotImplementedError

    ##############################################

    def __rmod__(self, other):
        """other % self"""
        raise NotImplementedError

    ##############################################

    def __lt__(self, other):
        """self < other

        < on Reals defines a total ordering, except perhaps for NaN.

        """
        return float(self) < float(other)

    ##############################################

    def __le__(self, other):
        """self <= other"""
        return float(self) <= float(other)

    ##############################################

    def __ceil__(self):
        return math.ceil(float(self))

    ##############################################

    def __floor__(self):
        return math.floor(float(self))

    ##############################################

    def __round__(self):
        return round(float(self))

    ##############################################

    def reciprocal(self):
        equivalent_unit = self.unit.reciprocal(prefixed_unit=True)
        reciprocal_value = 1. / float(self)
        return equivalent_unit.new_value(reciprocal_value)

    ##############################################

    def get_prefixed_unit(self, power=0):
        prefixed_unit = PrefixedUnit.from_prefixed_unit(self.unit, power)
        if prefixed_unit is not None:
            return prefixed_unit
        else:
            raise NameError("Prefixed unit not found for {} and power {}".format(self, power))

    ##############################################

    def convert(self, prefixed_unit):
        """Convert the value to another power."""
        self._prefixed_unit.check_unit(prefixed_unit)
        if self._prefixed_unit.is_same_power(prefixed_unit):
            return self
        else:
            value = float(self) / prefixed_unit.scale
            return prefixed_unit.new_value(value)

    ##############################################

    def convert_to_power(self, power=0):
        """Convert the value to another power."""
        if power == 0:
            value = float(self)
        else:
            value = float(self) / 10**power
        return self.get_prefixed_unit(power).new_value(value)

    ##############################################

    def canonise(self):

        # log10(10**n) = n    log10(1) = 0   log10(10**-n) = -n   log10(0) = -oo

        try:
            abs_value = abs(float(self))
            log = math.log(abs_value)/math.log(1000)
            # if abs_value >= 1:
            #     power = 3 * int(log)
            # else:
            #     if log - int(log): # frac
            #         power = 3 * (int(log) -1)
            #     else:
            #         power = 3 * int(log)
            power = int(log)
            if abs_value < 1 and (log - int(log)):
                power -= 1
            power *= 3
            # print('Unit.canonise', self, self._value, int(self._power), '->', float(self), power)
            if power == int(self.power):
                # print('Unit.canonise noting to do for', self)
                return self
            else:
                # print('Unit.canonise convert', self, 'to', power)
                # print('Unit.canonise convert', self, 'to', Unit)
                return self.convert_to_power(power)
        except Exception as e: # Fixme: fallback
            self._logger.warning(e)
            return self

####################################################################################################

class UnitValues(np.ndarray):

    """This class implements a Numpy array with a unit and a power (prefix).

    """

    _logger = _module_logger.getChild('UnitValues')

    CONVERSION = EnumFactory('ConversionType', (
        'NOT_IMPLEMENTED',
        'NO_CONVERSION',
        'FLOAT',
        'UNIT_MATCH',
        'UNIT_MATCH_NO_OUT_CAST',
        'NEW_UNIT'
        ))

    # Reference_documentation:
    #   https://docs.scipy.org/doc/numpy-1.13.0/reference/arrays.ndarray.html
    #   https://docs.scipy.org/doc/numpy-1.13.0/user/basics.subclassing.html
    #   https://docs.scipy.org/doc/numpy-1.13.0/reference/ufuncs.html
    UFUNC_MAP = {
        # Math operations
        # --------------------------------------------------
        np.add:          CONVERSION.UNIT_MATCH,
        np.subtract:     CONVERSION.UNIT_MATCH,
        np.multiply:     CONVERSION.NEW_UNIT,
        np.divide:       CONVERSION.NEW_UNIT,
        np.logaddexp:    CONVERSION.FLOAT,
        np.logaddexp2:   CONVERSION.FLOAT,
        np.true_divide:  CONVERSION.NEW_UNIT,
        np.floor_divide: CONVERSION.NEW_UNIT,
        np.negative:     CONVERSION.NO_CONVERSION,
        np.positive:     CONVERSION.NO_CONVERSION,
        np.power:        CONVERSION.NEW_UNIT,
        np.remainder:    CONVERSION.UNIT_MATCH,
        np.mod:          CONVERSION.UNIT_MATCH,
        np.fmod:         CONVERSION.UNIT_MATCH,
        np.divmod:       CONVERSION.UNIT_MATCH,
        np.absolute:     CONVERSION.NO_CONVERSION,
        np.fabs:         CONVERSION.NO_CONVERSION,
        np.rint:         CONVERSION.NO_CONVERSION,
        np.sign:         CONVERSION.NO_CONVERSION,
        np.heaviside:    CONVERSION.NOT_IMPLEMENTED, # !
        np.conj:         CONVERSION.NOT_IMPLEMENTED, # !
        np.exp:          CONVERSION.FLOAT,
        np.exp2:         CONVERSION.FLOAT,
        np.log:          CONVERSION.FLOAT,
        np.log2:         CONVERSION.FLOAT,
        np.log10:        CONVERSION.FLOAT,
        np.expm1:        CONVERSION.FLOAT,
        np.log1p:        CONVERSION.FLOAT,
        np.sqrt:         CONVERSION.NEW_UNIT,
        np.square:       CONVERSION.NEW_UNIT,
        np.cbrt:         CONVERSION.NEW_UNIT,
        np.reciprocal:   CONVERSION.NEW_UNIT,

        # Trigonometric functions
        # --------------------------------------------------
        np.sin:     CONVERSION.FLOAT,
        np.cos:     CONVERSION.FLOAT,
        np.tan:     CONVERSION.FLOAT,
        np.arcsin:  CONVERSION.FLOAT,
        np.arccos:  CONVERSION.FLOAT,
        np.arctan:  CONVERSION.FLOAT,
        np.arctan2: CONVERSION.FLOAT,
        np.hypot:   CONVERSION.FLOAT,
        np.sinh:    CONVERSION.FLOAT,
        np.cosh:    CONVERSION.FLOAT,
        np.tanh:    CONVERSION.FLOAT,
        np.arcsinh: CONVERSION.FLOAT,
        np.arccosh: CONVERSION.FLOAT,
        np.arctanh: CONVERSION.FLOAT,
        np.deg2rad: CONVERSION.FLOAT,
        np.rad2deg: CONVERSION.FLOAT,

        # Bit-twiddling functions
        # --------------------------------------------------
        np.bitwise_and: CONVERSION.NOT_IMPLEMENTED, # Nonsense
        np.bitwise_or:  CONVERSION.NOT_IMPLEMENTED, # Nonsense
        np.bitwise_xor: CONVERSION.NOT_IMPLEMENTED, # Nonsense
        np.invert:      CONVERSION.NOT_IMPLEMENTED, # Nonsense
        np.left_shift:  CONVERSION.NOT_IMPLEMENTED, # Nonsense
        np.right_shift: CONVERSION.NOT_IMPLEMENTED, # Nonsense

        # Comparison functions
        # --------------------------------------------------
        np.greater:       CONVERSION.UNIT_MATCH_NO_OUT_CAST,
        np.greater_equal: CONVERSION.UNIT_MATCH_NO_OUT_CAST,
        np.less:          CONVERSION.UNIT_MATCH_NO_OUT_CAST,
        np.less_equal:    CONVERSION.UNIT_MATCH_NO_OUT_CAST,
        np.not_equal:     CONVERSION.UNIT_MATCH_NO_OUT_CAST,
        np.equal:         CONVERSION.UNIT_MATCH_NO_OUT_CAST,

        np.logical_and:   CONVERSION.UNIT_MATCH,
        np.logical_or:    CONVERSION.UNIT_MATCH,
        np.logical_xor:   CONVERSION.UNIT_MATCH,
        np.logical_not:   CONVERSION.UNIT_MATCH,

        np.maximum:       CONVERSION.UNIT_MATCH,
        np.minimum:       CONVERSION.UNIT_MATCH,
        np.fmax:          CONVERSION.UNIT_MATCH,
        np.fmin:          CONVERSION.UNIT_MATCH,

        # Floating functions
        # --------------------------------------------------
        np.isfinite:  CONVERSION.NOT_IMPLEMENTED, # ! _T
        np.isinf:     CONVERSION.NOT_IMPLEMENTED, # ! _T
        np.isnan:     CONVERSION.NOT_IMPLEMENTED, # ! _T
        np.fabs:      CONVERSION.NOT_IMPLEMENTED, # ! _
        np.signbit:   CONVERSION.NOT_IMPLEMENTED, # ! _T
        np.copysign:  CONVERSION.NOT_IMPLEMENTED, # !
        np.nextafter: CONVERSION.NOT_IMPLEMENTED, # !
        np.spacing:   CONVERSION.NOT_IMPLEMENTED, # !
        np.modf:      CONVERSION.NOT_IMPLEMENTED, # !
        np.ldexp:     CONVERSION.NOT_IMPLEMENTED, # !
        np.frexp:     CONVERSION.NOT_IMPLEMENTED, # !
        np.fmod:      CONVERSION.NOT_IMPLEMENTED, # !
        np.floor:     CONVERSION.NOT_IMPLEMENTED, # !
        np.ceil:      CONVERSION.NO_CONVERSION,
        np.trunc:     CONVERSION.NO_CONVERSION,

        # Statistic functions
        # --------------------------------------------------
        np.mean:      CONVERSION.NO_CONVERSION,
    }

    ##############################################

    @classmethod
    def from_ndarray(cls, array, prefixed_unit):

        # cls._logger.debug('UnitValues.__new__ ' + str((cls, array, prefixed_unit)))

        # obj = cls(prefixed_unit, array.shape, array.dtype) # Fixme: buffer ???
        # obj[...] = array[...]

        obj = array.view(UnitValues)
        obj._prefixed_unit = prefixed_unit

        if isinstance(array, UnitValues):
            return array.convert(prefixed_unit)

        return obj

    ##############################################

    def __new__(cls,
                prefixed_unit,
                shape, dtype=float, buffer=None, offset=0, strides=None, order=None):

        # Called for explicit constructor
        #  obj = UnitValues(prefixed_unit, shape)

        # cls._logger.debug('UnitValues.__new__ ' + str((cls, prefixed_unit, shape, dtype, buffer, offset, strides, order)))

        obj = super(UnitValues, cls).__new__(cls, shape, dtype, buffer, offset, strides, order)
        # obj = np.asarray(input_array).view(cls)

        obj._prefixed_unit = prefixed_unit

        return obj

    ##############################################

    def __array_finalize__(self, obj):

        # self._logger.debug('UnitValues.__new__ ' + '\n  {}'.format(obj))

        # self is a new object resulting from ndarray.__new__(UnitValues, ...)
        # therefore it only has attributes that the ndarray.__new__ constructor gave it
        # i.e. those of a standard ndarray.

        # We could have got to the ndarray.__new__ call in 3 ways:

        # From an explicit constructor - e.g. UnitValues():
        #    obj is None
        #    we are in the middle of the UnitValues.__new__ constructor

        if obj is None:
            return

        # From view casting - e.g arr.view(UnitValues):
        #    obj is arr
        #    type(obj) can be UnitValues

        # From new-from-template - e.g infoarr[:3]
        #    type(obj) is UnitValues

        self._prefixed_unit = getattr(obj, '_prefixed_unit', None) # Fixme: None

    ##############################################

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):

        # - "ufunc" is the ufunc object that was called
        # - "method" is a string indicating how the ufunc was called, either
        #       "__call__" to indicate it was called directly,
        #       or one of its "ufuncs.methods": "reduce", "accumulate",  "reduceat", "outer", or "at".
        # - "inputs" is a tuple of the input arguments to the ufunc
        # - "kwargs" contains any optional or keyword arguments passed to the function.
        #    This includes any *out* arguments, which are always contained in a tuple.

        # ufunc.reduce(a[, axis, dtype, out, keepdims])     Reduces a‘s dimension by one, by applying ufunc along one axis.
        # ufunc.accumulate(array[, axis, dtype, out, ...])  Accumulate the result of applying the operator to all elements.
        # ufunc.reduceat(a, indices[, axis, dtype, out])    Performs a (local) reduce with specified slices over a single axis.
        # ufunc.outer(A, B, **kwargs)                       Apply the ufunc op to all pairs (a, b) with a in A and b in B.
        # ufunc.at(a, indices[, b])                         Performs unbuffered in place operation on operand ‘a’ for elements specified by ‘indices’.

        # self._logger.debug(
        #     '\n  self={}\n  ufunc={}\n  method={}\n  inputs={}\n  kwargs={}'
        #     .format(self, ufunc, method, inputs, kwargs))

        # ufunc=<ufunc 'multiply'>
        # method=__call__
        # inputs=(UnitValues(mV, [0 1 2 3 4 5 6 7 8 9]), 2)

        # ufunc=<ufunc 'sin'>
        # method=__call__
        # inputs=(UnitValues(mV, [0 1 2 3 4 5 6 7 8 9]),)
        # kwargs={}

        # ufunc=<ufunc 'add'>
        # method=__call__
        # inputs=(UnitValues(mV, [0 1 2 3 4 5 6 7 8 9]), UnitValues(mV, [0 1 2 3 4 5 6 7 8 9]))

        # ufunc=<ufunc 'add'>
        # method=reduce
        # inputs=(WaveForm  [10 12 14 16 18 20 22 24 26 28]@mV,)

        prefixed_unit = self._prefixed_unit

        conversion = self.UFUNC_MAP[ufunc]
        self._logger.debug("Conversion for {} is {}".format(ufunc, conversion))

        # e.g. np.mean do an internal call to reduce
        if method != '__call__':
            conversion = self.CONVERSION.NO_CONVERSION

        # Cast inputs to ndarray
        args = []
        if conversion == self.CONVERSION.NO_CONVERSION:
            # should be 1 arg
            args = [( input_.as_ndarray(False) if isinstance(input_, UnitValues) else input_ )
                    for input_ in inputs]
        #
        elif conversion == self.CONVERSION.FLOAT:
            if not prefixed_unit.is_unit_less:
                # raise ValueError("Must be unit less")
                self._logger.warning("Should be unit less")
            args = [( input_.as_ndarray(True) if isinstance(input_, UnitValues) else input_ )
                    for input_ in inputs]
        #
        elif conversion in (self.CONVERSION.UNIT_MATCH, self.CONVERSION.UNIT_MATCH_NO_OUT_CAST):
            # len(inputs) == 2
            other = inputs[1]
            if isinstance(other, (UnitValues, UnitValue)):
                self._check_unit(other)
                args.append(self.as_ndarray())
                nd_other = self._convert_value(other)
                if isinstance(other, UnitValues):
                    nd_other = nd_other.as_ndarray()
                elif isinstance(other, UnitValue):
                    nd_other = float(nd_other)
                args.append(nd_other)
            else:
                raise ValueError
        #
        elif conversion == self.CONVERSION.NEW_UNIT:
            if len(inputs) == 1:
                #! Fixme: power
                if ufunc == np.sqrt:
                    prefixed_unit = self.unit.sqrt(True)
                elif ufunc == np.square:
                    prefixed_unit = self.unit.square(True)
                elif ufunc == np.cbrt:
                    prefixed_unit = self.unit.cbrt(True)
                elif ufunc == np.reciprocal:
                    prefixed_unit = self.unit.reciprocal(True)
                else:
                    raise NotImplementedError
                args.append(self.as_ndarray(True))
            elif len(inputs) == 2:
                other = inputs[1]
                if isinstance(other, (UnitValues, UnitValue)):
                    if ufunc == np.multiply:
                        prefixed_unit = self.unit.multiply(other.unit, True)
                    elif ufunc in (np.divide, np.true_divide, np.floor_divide):
                        prefixed_unit = self.unit.divide(other.unit, True)
                    else:
                        raise NotImplementedError
                    args.append(self.as_ndarray(True))
                    if isinstance(other, UnitValue):
                        args.append(float(other))
                    else:
                        args.append(other.as_ndarray(True))
                elif ufunc in (np.multiply, np.divide, np.true_divide, np.floor_divide, np.power):
                    if ufunc == np.power:
                        prefixed_unit = self.unit.power(other, True)
                    args.append(self.as_ndarray())
                    args.append(other)
                else:
                    raise NotImplementedError
            else:
                raise NotImplementedError
        #
        else: # self.CONVERSION.NOT_IMPLEMENTED
            raise NotImplementedError

        # self._logger.debug("Output unit is {}".format(prefixed_unit))

        # Cast outputs to ndarray
        outputs = kwargs.pop('out', None)
        if outputs:
            out_args = []
            for output in outputs:
                if isinstance(output, UnitValues):
                    out_args.append(output.as_ndarray())
                else:
                    out_args.append(output)
            kwargs['out'] = tuple(out_args)
        else:
            outputs = (None,) * ufunc.nout

        # Call ufunc
        results = super(UnitValues, self).__array_ufunc__(ufunc, method, *args, **kwargs)
        if results is NotImplemented:
            return NotImplemented

        # ensure results is a tuple
        if ufunc.nout == 1:
            results = (results,)

        # Cast results
        if conversion in (self.CONVERSION.FLOAT, self.CONVERSION.UNIT_MATCH_NO_OUT_CAST):
            # Fixme: ok ???
            results = tuple(( result if output is None else output )
                            for result, output in zip(results, outputs))
        else:
            results = tuple(( UnitValues.from_ndarray(np.asarray(result), prefixed_unit) if output is None else output )
                            for result, output in zip(results, outputs))

        # list or scalar
        return results[0] if len(results) == 1 else results

    ##############################################

#   def __array_wrap__(self, out_array, context=None):
#
#       self._logger.debug('\n  self={}\n  out_array={}\n  context={}'.format(self, out_array, context))
#
#       return super(UnitValues, self).__array_wrap__(out_array, context)

    ##############################################

    def as_ndarray(self, scale=False):
        array = self.view(np.ndarray)
        if scale:
            return array * self.scale
        else:
            return array

    ##############################################

    def __getitem__(self, slice_):

        value = super(UnitValues, self).__getitem__(slice_)

        if isinstance(value, UnitValue): # slice
            return value
        else:
            return self._prefixed_unit.new_value(value)

    ##############################################

    def __setitem__(self, slice_, value):

        if isinstance(value, UnitValue):
            self._check_unit(value)
            value = self._convert_value(value).value
        elif isinstance(value, UnitValues):
            self._check_unit(value)
            value = self._convert_value(value)

        super(UnitValues, self).__setitem__(slice_, value)

    ##############################################

    # def __getstate__(self):
    #     # https://docs.python.org/3/library/pickle.html#object.__getstate__
    #     return {
    #         'data': super(UnitValues, self).__getstate__(),
    #         'prefixed_unit': self._prefixed_unit,
    #     }

    ##############################################

    def __reduce__(self):
        # https://docs.python.org/3/library/pickle.html#object.__reduce__
        np_state = super(UnitValues, self).__reduce__()
        # ( <built-in function _reconstruct>,
        #   (<class 'PySpice.Unit.Unit.UnitValues'>, (0,), b'b'),
        #   (1, (1, 1), dtype('float64'), False, b'\x00\x00\x80?\x00\x00\x80?') )
        obj_state = (self._prefixed_unit,) + np_state[2]
        return np_state[:2] + (obj_state,) + np_state[3:]

    ##############################################

    def __setstate__(self, state):
        # https://docs.python.org/3/library/pickle.html#object.__setstate__
        super(UnitValues, self).__setstate__(state[1:])
        self._prefixed_unit = state[0]

    ##############################################

    def __contains__(self, value):
        raise NotImplementedError

    ##############################################

    def __repr__(self):
        # return repr(self.as_ndarray())
        return '{}({})'.format(self.__class__.__name__, str(self))

    ##############################################

    @property
    def prefixed_unit(self):
        return self._prefixed_unit

    @property
    def unit(self):
        return self._prefixed_unit.unit

    @property
    def power(self):
        return self._prefixed_unit.power

    @property
    def scale(self):
        return self._prefixed_unit.power.scale

    ##############################################

    def is_same_unit(self, other):
        return self._prefixed_unit.is_same_unit(other.prefixed_unit)

    ##############################################

    def _check_unit(self, other):
        if not self.is_same_unit(other):
            raise UnitError

    ##############################################

    def is_same_power(self, other):
        return self._prefixed_unit.is_same_power(other.prefixed_unit)

    ##############################################

    def __eq__(self, other):
        """self == other"""
        if isinstance(other, UnitValues):
            return self.is_same_unit(other) and self.as_ndarray() == other.as_ndarray()
        else:
            raise ValueError

    ##############################################

    def _convert_value(self, other):
        """Convert the value of other to the power of self."""
        self._check_unit(other)
        if self.is_same_power(other):
            return other
        else:
            return other * (other.scale / self.scale) # for numerical precision

    ##############################################

    def __str__(self):
        return str(self.as_ndarray()) + '@' + str(self._prefixed_unit)

    ##############################################

    def reciprocal(self):
        equivalent_unit = self.unit.reciprocal(prefixed_unit=True)
        reciprocal_value = 1. / np.as_ndarray(True)
        return self.from_ndarray(reciprocal_value, equivalent_unit)

    ##############################################

    def get_prefixed_unit(self, power=0):
        prefixed_unit = PrefixedUnit.from_prefixed_unit(self.unit, power)
        if prefixed_unit is not None:
            return prefixed_unit
        else:
            raise NameError("Prefixed unit not found for {} and power {}".format(self, power))

    ##############################################

    def convert(self, prefixed_unit):
        """Convert the value to another power."""
        self._prefixed_unit.check_unit(prefixed_unit)
        if self._prefixed_unit.is_same_power(prefixed_unit):
            return self
        else:
            value = self.as_ndarray(True) / prefixed_unit.scale
            return prefixed_unit.new_value(value)

    ##############################################

    def convert_to_power(self, power=0):
        """Convert the value to another power."""
        value = self.as_ndarray(True)
        if power != 0:
            value /= 10**power
        return self.get_prefixed_unit(power).new_value(value)

####################################################################################################

# Reset
PrefixedUnit._value_ctor = UnitValue
PrefixedUnit._values_ctor = UnitValues

_simple_prefixed_unit = PrefixedUnit()

####################################################################################################

class FrequencyMixin:

    """ This class implements a frequency mixin. """

    ##############################################

    @property
    def period(self):
        r""" Return the period :math:`T = \frac{1}{f}`. """
        return self.reciprocal()

    ##############################################

    @property
    def pulsation(self):
        r""" Return the pulsation :math:`\omega = 2\pi f`. """
        # Fixme: UnitValues
        return float(self * 2 * math.pi)

####################################################################################################

class PeriodMixin:

    """ This class implements a period mixin. """

    ##############################################

    @property
    def frequency(self):
        r""" Return the period :math:`f = \frac{1}{T}`. """
        return self.reciprocal()

    ##############################################

    @property
    def pulsation(self):
        r""" Return the pulsation :math:`\omega = \frac{2\pi}{T}`. """
        return self.frequency.pulsation



#r# This example shows how to simulate and plot the characteristic curve of a diode.
#r#
#r# Theory
#r# ------

#r# A diode is a semiconductor device made of a PN junction which is a sandwich of two doped silicon
#r# layers.

#f# tikz('diode.tex', width=500)

#r# Before two explains the purpose of the silicon doping.  We will give a quick and simplified look
#r# on the atomic world.  An atom is made of the same number of protons and electrons, this number
#r# is so called Z and characterise the atom.  Each electron is coupled to the proton's kernel by
#r# the electromagnetic interaction, but with different levels of energy.  The reason is due to the
#r# fact each electron screens each other.  There is thus some electrons which are strongly coupled
#r# and other ones which are weakly coupled, so called valence's electrons.  An atom is a neutral
#r# object when we observe it at a large distance, but when the pressure and temperature of the
#r# environment match some conditions, the weakly coupled electrons can be shared with atoms in the
#r# neighbourhood and create the electromagnetic interaction between atoms which make the cohesion
#r# of the matter.

#r#
#r# Depending on the weakness of the electrons, an atom can be an insulated material or a conductor.
#r# Semiconductors are between them.

#r#
#r# The doping consists to diffuse a small quantities of atoms with a larger or smaller number of
#r# valence electron in a silicon layer.  Since a silicon atom has four valence electrons, we use
#r# atoms having 3 or 5 valence electrons for the doping.  A doping using a larger number of valence
#r# electrons is called N for negative, and P respectively.  In a silicon lattice doped with atoms
#r# having a larger number of valence electrons, the additional electrons do not participate to the
#r# lattice cohesion and are weakly coupled to the atoms.  The conductivity is thus improved.  In
#r# other hands, a silicon lattice doped with atoms having a smaller number of valence electrons,
#r# some electrons are missing for the lattice cohesion.  These missing electrons are called holes
#r# since the P doping atoms will catch free electrons so as to normalise the lattice cohesion.

#r#
#r# When we make a sandwich of P and N doped silicon layers, the weak electrons of the N layer
#r# diffuse to the P layer until an equilibrium state is achieved.  Indeed this diffusion creates a
#r# depletion region which acts as a barrier to the electrons, since a P doping atom becomes negative
#r# (anion) when it catches an electron and reciprocally a N doping atom becomes positive (cation)
#r# when it lost an electron.  The depletion region is thus a kind of capacitor.

#r#
#r# The volume of the depletion region can be changed by applying a tension across the PN junction.
#r# If the tension between the PN junction is negative then the depletion region is enlarged, and
#r# only a very small current can flow through the junction due to the thermal agitation.  But if we
#r# apply a positive tension, the depletion region is pressurised, and if it reaches a threshold, an
#r# electron flow will be able to pass through the junction.

#r#
#r# A PN junction can thus only conduct a current from the anode to the cathode and only if a
#r# minimal bias tension is applied across it.

#r#
#r# However if a large enough inverse tension is applied to the junction then the electrostatic
#r# force will become sufficiently large enough to pull off electrons across the junction and the
#r# current will flow from the cathode to the anode. This effect is called breakdown.

#r#
#r# Simulation
#r# ----------

####################################################################################################

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Doc.ExampleTools import find_libraries
from PySpice.Spice.Netlist import Circuit
from PySpice.Spice.Library import SpiceLibrary
from PySpice.Unit import *
from PySpice.Physics.SemiConductor import ShockleyDiode

####################################################################################################

libraries_path = find_libraries()
spice_library = SpiceLibrary(libraries_path)

####################################################################################################

#r# For this purpose, we use the common high-speed diode 1N4148.  The diode is driven by a variable
#r# voltage source through a limiting current resistance.

#f# circuit_macros('diode-characteristic-curve-circuit.m4')

circuit = Circuit('Diode Characteristic Curve')

circuit.include(spice_library['1N4148'])

circuit.V('input', 'in', circuit.gnd, 10@u_V)
circuit.R(1, 'in', 'out', 1@u_Ω) # not required for simulation
circuit.X('D1', '1N4148', 'out', circuit.gnd)

#r# We simulate the circuit at these temperatures: 0, 25 and 100 °C.

# Fixme: Xyce ???
temperatures = [0, 25, 100]@u_Degree
analyses = {}
for temperature in temperatures:
    simulator = circuit.simulator(temperature=temperature, nominal_temperature=temperature)
    analysis = simulator.dc(Vinput=slice(-2, 5, .01))
    analyses[float(temperature)] = analysis

####################################################################################################

#r# We plot the characteristic curve and compare it to the Shockley diode model:
#r#
#r# .. math::
#r#
#r#     I_d = I_s \left( e^{\frac{V_d}{n V_T}} - 1 \right)
#r#
#r# where :math:`V_T = \frac{k T}{q}`
#r#
#r# In order to scale the reverse biased region, we have to do some hack with Matplotlib.
#r#

silicon_forward_voltage_threshold = .7

shockley_diode = ShockleyDiode(Is=4e-9, degree=25)

def two_scales_tick_formatter(value, position):
    if value >= 0:
        return '{} mA'.format(value)
    else:
        return '{} nA'.format(value/100)
formatter = ticker.FuncFormatter(two_scales_tick_formatter)

figure, (ax1, ax2) = plt.subplots(2, figsize=(20, 10))

ax1.set_title('1N4148 Characteristic Curve ')
ax1.set_xlabel('Voltage [V]')
ax1.set_ylabel('Current')
ax1.grid()
ax1.set_xlim(-2, 2)
ax1.axvspan(-2, 0, facecolor='green', alpha=.2)
ax1.axvspan(0, silicon_forward_voltage_threshold, facecolor='blue', alpha=.1)
ax1.axvspan(silicon_forward_voltage_threshold, 2, facecolor='blue', alpha=.2)
ax1.set_ylim(-500, 750) # Fixme: round
ax1.yaxis.set_major_formatter(formatter)
Vd = analyses[25].out
# compute scale for reverse and forward region
forward_region = Vd >= 0@u_V
reverse_region = np.invert(forward_region)
scale =  reverse_region*1e11 + forward_region*1e3
#?# check temperature
for temperature in temperatures:
    analysis = analyses[float(temperature)]
    ax1.plot(Vd, - analysis.Vinput * scale)
ax1.plot(Vd, shockley_diode.I(Vd) * scale, 'black')
ax1.legend(['@ {} °C'.format(temperature)
            for temperature in temperatures] + ['Shockley Diode Model Is = 4 nA'],
           loc=(.02,.8))
ax1.axvline(x=0, color='black')
ax1.axhline(y=0, color='black')
ax1.axvline(x=silicon_forward_voltage_threshold, color='red')
ax1.text(-1, -100, 'Reverse Biased Region', ha='center', va='center')
ax1.text( 1, -100, 'Forward Biased Region', ha='center', va='center')

#r# Now we compute and plot the static and dynamic resistance.
#r#
#r# .. math::
#r#
#r#   \frac{d I_d}{d V_d} = \frac{1}{n V_T}(I_d + I_s)
#r#
#r# .. math::
#r#
#r#   r_d = \frac{d V_d}{d I_d} \approx \frac{n V_T}{I_d}

ax2.set_title('Resistance @ 25 °C')
ax2.grid()
ax2.set_xlim(-2, 3)
ax2.axvspan(-2, 0, facecolor='green', alpha=.2)
ax2.axvspan(0, silicon_forward_voltage_threshold, facecolor='blue', alpha=.1)
ax2.axvspan(silicon_forward_voltage_threshold, 3, facecolor='blue', alpha=.2)
analysis = analyses[25]
static_resistance = -analysis.out / analysis.Vinput
dynamic_resistance = np.diff(-analysis.out) / np.diff(analysis.Vinput)
ax2.semilogy(analysis.out, static_resistance, base=10)
ax2.semilogy(analysis.out[10:-1], dynamic_resistance[10:], base=10)
ax2.axvline(x=0, color='black')
ax2.axvline(x=silicon_forward_voltage_threshold, color='red')
ax2.axhline(y=1, color='red')
ax2.text(-1.5, 1.1, 'R limitation = 1 Ω', color='red')
ax2.legend(['{} Resistance'.format(x) for x in ('Static', 'Dynamic')], loc=(.05,.2))
ax2.set_xlabel('Voltage [V]')
ax2.set_ylabel('Resistance [Ω]')

plt.tight_layout()
plt.show()

#f# save_figure('figure', 'diode-characteristic-curve.png')

#r# We observe the forward voltage threshold increase with the temperature.
#?# #r# Indeed the thermal agitation overcome the electron flow.



print("import: 'PySpice'")
import PySpice

print("import: 'PySpice.Config'")
import PySpice.Config

print("import: 'PySpice.Doc'")
import PySpice.Doc

print("import: 'PySpice.Logging'")
import PySpice.Logging

print("import: 'PySpice.Math'")
import PySpice.Math

print("import: 'PySpice.Physics'")
import PySpice.Physics

print("import: 'PySpice.Plot'")
import PySpice.Plot

print("import: 'PySpice.Probe'")
import PySpice.Probe

print("import: 'PySpice.Spice'")
import PySpice.Spice

print("import: 'PySpice.Tools'")
import PySpice.Tools

print("import: 'PySpice.Unit'")
import PySpice.Unit




####################################################################################################

#r# =============================
#r#  Voltage and Current Divider
#r# =============================

#r# This circuit is a fundamental block in electronic that permits to scale a voltage by an
#r# impedance ratio:

#f# circuit_macros('voltage-divider.m4')

#r# The relation between the input and ouput voltage is:
#r#
#r# .. math::
#r#
#r#     \frac{V_{out}}{V_{in}} = \frac{R_2}{R_1 + R_2}
#r#
#r# This equation holds for any impedances like resistance, capacitance, inductance, etc.

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Spice.Netlist import Circuit
from PySpice.Unit import *

####################################################################################################

circuit = Circuit('Voltage Divider')

circuit.V('input', 1, circuit.gnd, 10@u_V)
circuit.R(1, 1, 2, 2@u_kΩ)
circuit.R(2, 2, circuit.gnd, 1@u_kΩ)

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.operating_point()

for node in analysis.nodes.values():
    print('Node {}: {:5.2f} V'.format(str(node), float(node))) # Fixme: format value + unit
#o#

####################################################################################################

#r# Similarly we can build a circuit that scale a current by an impedance ratio:

#f# circuit_macros('current-divider.m4')

#r# The relation between the input and ouput current is:
#r#
#r# .. math::
#r#
#r#     \frac{I_{out}}{I_{in}} = \frac{R_1}{R_1 + R_2}
#r#
#r# Note the role of R1 and R2 is exchanged.
#r#
#r# This equation holds for any impedances like resistance, capacitance, inductance, etc.

####################################################################################################

circuit = Circuit('Current Divider')

circuit.I('input', 1, circuit.gnd, 1@u_A) # Fixme: current value
circuit.R(1, 1, circuit.gnd, 2@u_kΩ)
circuit.R(2, 1, circuit.gnd, 1@u_kΩ)

for resistance in (circuit.R1, circuit.R2):
    resistance.minus.add_current_probe(circuit) # to get positive value

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.operating_point()

# Fixme: current over resistor
for node in analysis.branches.values():
    print('Node {}: {:5.2f} A'.format(str(node), float(node))) # Fixme: format value + unit
#o#
